import numpy as np
import astropy.units as u
import pandas as pd

from cosmic.sample.initialbinarytable import InitialBinaryTable

from ..pop import Population
from ..sfh import StarFormationHistory

from .utils import dispersion_from_virial_parameter
import warnings


__all__ = ["HydroPopulation"]


class HydroPopulation(Population):
    def __init__(self, star_particles, cluster_radius=3 * u.pc, cluster_mass=1e4 * u.Msun,
                 virial_parameter=1.0, subset=None, sampling_params={}, snapshot_type=None, **kwargs):
        """A population of stars sampled from a hydrodynamical zoom-in snapshot

        Each star particle in the snapshot is converted to a (binary) stellar population, sampled with COSMIC,
        with initial positions and kinematics determined based on the virial parameter and particle size.

        Parameters
        ----------
        star_particles : :class:`~pandas.DataFrame`
            A table of star particles from a snapshot that contains the mass, metallicity, formation time,
            position and velocity of each particle, as returned by
            :func:`~cogsworth.hydro.rewind.rewind_to_formation`
        cluster_radius : :class:`~astropy.units.Quantity`, optional
            Size of gaussian for cluster for each star particle, by default 3 * u.pc
        cluster_mass : :class:`~astropy.units.Quantity`, optional
            Mass of cluster for each star particle for calculating velocity dispersion,
            by default 1e4 * u.Msun
        virial_parameter : `float`, optional
            Virial parameter for each cluster, for setting velocity dispersions, by default 1.0
        subset : `int` or `list` of `int`, optional
            If set, only use a subset of the star particles, by default use all of them. If an integer is
            passed, a random subset of that size is used, if a list of integers is passed, only those IDs are
            used.
        sampling_params : `dict`, optional
            As in :class:`~cogsworth.pop.Population` BUT changing the defaults to
            {"sampling_target": "total_mass", "trim_extra_samples": True} (i.e. to sample to the total mass of
            the star particle)
        snapshot_type : `str`, optional
            What sort of snapshot this to infer what citations to use, by default None (i.e. don't infer).
            Currently supported: "FIRE"
        **kwargs : various
            Parameters to pass to :class:`~cogsworth.pop.Population`
        """
        self.star_particles = star_particles
        self.cluster_radius = cluster_radius
        self.cluster_mass = cluster_mass
        self.virial_parameter = virial_parameter
        self._subset_inds = self.star_particles.index.values
        if subset is not None and isinstance(subset, int):
            self._subset_inds = np.random.choice(self._subset_inds, size=subset, replace=False)
        elif subset is not None:
            self._subset_inds = subset
        if "n_binaries" not in kwargs:
            kwargs["n_binaries"] = None

        base_sampling_params = {"sampling_target": "total_mass", "trim_extra_samples": True}
        base_sampling_params.update(sampling_params)
        super().__init__(sampling_params=base_sampling_params, **kwargs)

        if snapshot_type is not None:
            if snapshot_type.lower() == "fire":
                self.__citations__.append("FIRE")
        self.__citations__.append("pynbody")

    def __repr__(self):
        if self._orbits is None:
            return (f"<{self.__class__.__name__} - {len(self._subset_inds)} star particles - "
                    f"galactic_potential={self.galactic_potential.__class__.__name__}, "
                    f"SFH={self.sfh_model.__name__}>")
        else:
            return (f"<{self.__class__.__name__} - {len(self._subset_inds)} star particles - "
                    f"{self.n_binaries_match} evolved systems - "
                    f"galactic_potential={self.galactic_potential.__class__.__name__}, "
                    f"SFH={self.sfh_model.__name__}>")

    def __getitem__(self, ind):
        # convert any Pandas Series to numpy arrays
        ind = ind.values if isinstance(ind, pd.Series) else ind

        # ensure indexing with the right type
        ALLOWED_TYPES = (int, slice, list, np.ndarray, tuple)
        if not isinstance(ind, ALLOWED_TYPES):          # pragma: no cover
            raise ValueError((f"Can only index using one of {[at.__name__ for at in ALLOWED_TYPES]}, "
                              f"you supplied a '{type(ind).__name__}'"))

        # check validity of indices for array-like types
        if isinstance(ind, (list, tuple, np.ndarray)):
            # check every element is a boolean (if so, convert to bin_nums after asserting length sensible)
            if all(isinstance(x, (bool, np.bool_)) for x in ind):           # pragma: no cover
                assert len(ind) == len(self.bin_nums), "Boolean mask must be same length as the population"
                ind = self.bin_nums[ind]
            # otherwise ensure all elements are integers
            else:
                assert all(isinstance(x, (int, np.integer)) for x in ind), \
                    "Can only index using integers or a boolean mask"
                if len(np.unique(ind)) < len(ind):
                    warnings.warn(("You have supplied duplicate indices, this will invalidate the "
                                   "normalisation of the Population (e.g. mass_binaries will be wrong)"))

        # set up the bin_nums we are selecting
        bin_nums = ind

        # turn ints into arrays and convert slices to exact bin_nums
        if isinstance(ind, int):
            bin_nums = [ind]
        elif isinstance(ind, slice):
            bin_nums = self.bin_nums[ind]
        bin_nums = np.asarray(bin_nums)

        # check that the bin_nums are all valid
        check_nums = np.isin(bin_nums, self.bin_nums)
        if not check_nums.all():            # pragma: no cover
            raise ValueError(("The index that you supplied includes a `bin_num` that does not exist. "
                              f"The first bin_num I couldn't find was {bin_nums[~check_nums][0]}"))

        # start a new population with the same parameters
        new_pop = self.__class__(star_particles=self.star_particles, processes=self.processes,
                                 final_kstar1=self.final_kstar1,
                                 final_kstar2=self.final_kstar2, sfh_model=self.sfh_model,
                                 sfh_params=self.sfh_params, galactic_potential=self.galactic_potential,
                                 v_dispersion=self.v_dispersion, max_ev_time=self.max_ev_time,
                                 timestep_size=self.timestep_size, BSE_settings=self.BSE_settings,
                                 sampling_params=self.sampling_params,
                                 store_entire_orbits=self.store_entire_orbits,
                                 virial_parameter=self.virial_parameter, cluster_radius=self.cluster_radius)

        new_pop.n_binaries = len(bin_nums)
        new_pop.n_binaries_match = len(bin_nums)

        # proxy for checking whether sampling has been done
        if self._mass_binaries is not None:
            new_pop._mass_binaries = self._mass_binaries
            new_pop._mass_singles = self._mass_singles
            new_pop._n_singles_req = self._n_singles_req
            new_pop._n_bin_req = self._n_bin_req

        bin_num_to_ind = {num: i for i, num in enumerate(self.bin_nums)}
        sort_idx = np.argsort(list(bin_num_to_ind.keys()))
        idx = np.searchsorted(list(bin_num_to_ind.keys()), bin_nums, sorter=sort_idx)
        inds = np.asarray(list(bin_num_to_ind.values()))[sort_idx][idx]

        disrupted_bin_num_to_ind = {num: i for i, num in enumerate(self.bin_nums[self.disrupted])}
        sort_idx = np.argsort(list(disrupted_bin_num_to_ind.keys()))
        idx = np.searchsorted(list(disrupted_bin_num_to_ind.keys()),
                              bin_nums[np.isin(bin_nums, self.bin_nums[self.disrupted])],
                              sorter=sort_idx)
        inds_with_disruptions = np.asarray(list(disrupted_bin_num_to_ind.values()))[sort_idx][idx] + len(self)
        all_inds = np.concatenate((inds, inds_with_disruptions)).astype(int)

        if self._initial_galaxy is not None:
            new_pop._initial_galaxy = self._initial_galaxy[inds]

        # checking whether stellar evolution has been done
        if self._bpp is not None:
            # copy over subsets of data when they aren't None
            new_pop._bpp = self._bpp.loc[bin_nums]
            if self._bcm is not None:                   # pragma: no cover
                new_pop._bcm = self._bcm.loc[bin_nums]
            if self._initC is not None:
                new_pop._initC = self._initC.loc[bin_nums]
            if self._kick_info is not None:
                new_pop._kick_info = self._kick_info.loc[bin_nums]
            if self._final_bpp is not None:
                new_pop._final_bpp = self._final_bpp.loc[bin_nums]
            if self._disrupted is not None:
                new_pop._disrupted = self._disrupted[inds]
            if self._classes is not None:               # pragma: no cover
                new_pop._classes = self._classes.iloc[inds]
            if self._observables is not None:           # pragma: no cover
                new_pop._observables = self._observables.iloc[inds]

            # same thing but for arrays with appended disrupted secondaries
            if self._orbits is not None:
                new_pop._orbits = self.orbits[all_inds]
            if self._final_pos is not None:
                new_pop._final_pos = self._final_pos[all_inds]
            if self._final_vel is not None:
                new_pop._final_vel = self._final_vel[all_inds]
        return new_pop

    def get_citations(self, filename=None):
        super().get_citations(filename)
        print("\nNOTE: This population was sampled from a hydrodynamical zoom-in snapshot...but I don't know "
              "which one so I can't automate this citation, please cite the relevant paper(s) (e.g. for FIRE "
              "snapshots see here: http://flathub.flatironinstitute.org/fire)")

    def sample_initial_binaries(self):
        """Sample initial binaries from the star particles in the snapshot"""
        initial_binaries_list = [None for _ in range(len(self.star_particles))]
        self._mass_singles, self._mass_binaries, self._n_singles_req, self._n_bin_req = 0.0, 0.0, 0, 0

        for ibl_ind, i in enumerate(self._subset_inds):
            particle = self.star_particles.loc[i]
            samples = InitialBinaryTable.sampler('independent', self.final_kstar1, self.final_kstar2,
                                                 SF_start=(self.max_ev_time - particle["t_form"] * u.Gyr).to(u.Myr).value,
                                                 SF_duration=0.0, met=particle["Z"],
                                                 total_mass=particle["mass"],
                                                 size=particle["mass"] * 0.8,
                                                 **self.sampling_params)

            # apply the mass cutoff and record particle ID
            samples[0].reset_index(inplace=True)
            samples[0]["particle_id"] = np.repeat(i, len(samples[0]))

            # save samples
            initial_binaries_list[ibl_ind] = samples[0]
            self._mass_singles += samples[1]
            self._mass_binaries += samples[2]
            self._n_singles_req += samples[3]
            self._n_bin_req += samples[4]

        self._initial_binaries = pd.concat(initial_binaries_list, ignore_index=True)

        # these are the same for this class
        self.n_binaries = len(self._initial_binaries)
        self.n_binaries_match = len(self._initial_binaries)

        # ensure metallicities remain in a range valid for COSMIC
        self._initial_binaries.loc[self._initial_binaries["metallicity"] < 1e-4, "metallicity"] = 1e-4
        self._initial_binaries.loc[self._initial_binaries["metallicity"] > 0.03, "metallicity"] = 0.03

        self.sample_initial_galaxy()

    def sample_initial_galaxy(self):
        inds = np.searchsorted(self._subset_inds, self._initial_binaries["particle_id"].values)
        particles = self.star_particles.loc[self._subset_inds[inds]]
        x, y, z = particles["x"].values * u.kpc, particles["y"].values * u.kpc, particles["z"].values * u.kpc
        v_x = particles["v_x"].values * u.km / u.s
        v_y = particles["v_y"].values * u.km / u.s
        v_z = particles["v_z"].values * u.km / u.s

        pos = np.random.normal([x.to(u.kpc).value, y.to(u.kpc).value, z.to(u.kpc).value],
                               self.cluster_radius.to(u.kpc).value / np.sqrt(3),
                               size=(3, self.n_binaries_match)) * u.kpc

        v_R = (x * v_x + y * v_y) / (x**2 + y**2)**0.5
        v_T = (x * v_y - y * v_x) / (x**2 + y**2)**0.5

        vel_units = u.km / u.s
        dispersion = dispersion_from_virial_parameter(self.virial_parameter,
                                                      self.cluster_radius, self.cluster_mass)
        v_R, v_T, v_z = np.random.normal([v_R.to(vel_units).value,
                                          v_T.to(vel_units).value,
                                          v_z.to(vel_units).value],
                                         dispersion.to(vel_units).value / np.sqrt(3),
                                         size=(3, self.n_binaries_match)) * vel_units

        self._initial_galaxy = StarFormationHistory(self.n_binaries_match, immediately_sample=False)
        self._initial_galaxy._tau = self._initial_binaries["tphysf"].values * u.Myr
        self._initial_galaxy._Z = self._initial_binaries["metallicity"].values
        self._initial_galaxy._which_comp = np.repeat("FIRE", len(self.initial_galaxy._tau))

        self._initial_galaxy._x = pos[0]
        self._initial_galaxy._y = pos[1]
        self._initial_galaxy._z = pos[2]
        self._initial_galaxy.v_R = v_R
        self._initial_galaxy.v_T = v_T
        self._initial_galaxy.v_z = v_z

    def perform_stellar_evolution(self, **kwargs):
        """Perform stellar evolution on systems sampled from the star particles
        and track their parent particles"""
        super().perform_stellar_evolution(**kwargs)
        self._initC["particle_id"] = self._initial_binaries["particle_id"]



import unittest
import cogsworth
import pandas as pd


class Test(unittest.TestCase):
    def test_various_events(self):
        """Ensure functions work for different kinds of events

        I'm going to use fake bpp and kick_info tables for this. There will be four binaries
            - One with no events
            - One with a bound binary and a kick
            - One with a disruption from the first SN
            - One with a disruption from the second SN
        """

        bpp_dict = {
            "mass_1": [1, 10, 20, 1.4, 20, 20, 8],
            "mass_2": [1, 10, 20, 20, 20, 20, 1.4],
            "tphys": [0, 0, 0, 1, 0, 1, 2],
            "sep": [10, 10, 10, -1.0, 10, 10, -1.0],
            "ecc": [0, 0, 0, -1.0, 0, 0.1, -1.0],
            "evol_type": [-1, 15, 15, 11, 15, 16, 11],
            "bin_num": [0, 1, 2, 2, 3, 3, 3],
        }
        bpp = pd.DataFrame(data=bpp_dict)
        bpp.index = bpp["bin_num"].values

        kick_info_dict = {
            "star": [0, 1, 1, 1, 2],
            "disrupted": [0, 0, 1, 0, 1],
            "delta_vsysx_1": [0, 0, 0, 0, 0],
            "delta_vsysy_1": [0, 0, 0, 0, 0],
            "delta_vsysz_1": [0, 0, 0, 0, 0],
            "delta_vsysx_2": [0, 0, 0, 0, 0],
            "delta_vsysy_2": [0, 0, 0, 0, 0],
            "delta_vsysz_2": [0, 0, 0, 0, 0],
            "bin_num": [0, 1, 2, 3, 3],
        }
        kick_info = pd.DataFrame(data=kick_info_dict)
        kick_info.index = kick_info["bin_num"].values

        p = cogsworth.pop.Population(4, use_default_BSE_settings=True)
        p._bpp = bpp
        p._kick_info = kick_info
        p._initC = pd.DataFrame(
            data={"metallicity": [1e-2, 1e-2, 1e-2, 1e-2],
                  "bin_num": [0, 1, 2, 3]}
        )
        p._initC.index = p._initC["bin_num"].values
        p.final_bpp

        primary_events, secondary_events = cogsworth.events.identify_events(p)

        # 4 supernovae occurred
        self.assertTrue(len(primary_events) == 4)

        # 3 supernovae occurred in bound binaries
        self.assertTrue(len(secondary_events) == 3)

        # the first binary had no events so should not be present in either
        self.assertTrue(0 not in primary_events.index)
        self.assertTrue(0 not in secondary_events.index)

        # the secondary binary didn't disrupt, so shouldn't be in the secondary events table
        self.assertTrue(1 not in secondary_events.index)
        
        # the last binary had two supernovae so should have two events in the secondary table
        self.assertTrue(len(secondary_events.loc[[3]]) > len(secondary_events.loc[[2]]))



import numpy as np
import unittest
import cogsworth.obs.mist as mist
import astropy.units as u


class Test(unittest.TestCase):
    def test_bad_band(self):
        """Test that it breaks when given a bad band"""
        it_worked = True
        try:
            grid = mist.MISTBolometricCorrectionGrid(bands=("NOT A BAND",))
        except KeyError:
            it_worked = False
        self.assertFalse(it_worked)

    def test_build(self):
        """Test building and loading the HDF5 files"""
        grid = mist.MISTBolometricCorrectionGrid(bands=("Gaia_G_EDR3",), rebuild=True)
        h5_path = grid.build_hdf5("UBVRIplus")
        self.assertTrue(h5_path.exists())

        grid = mist.MISTBolometricCorrectionGrid(bands=("Gaia_G_EDR3",), rebuild=False)
        grid.download_filter_set("UBVRIplus")
        grid.extract_filter_set("UBVRIplus")
        grid.build_hdf5("UBVRIplus")

        self.assertTrue(h5_path.exists())

    def test_interpolator(self):
        """Test that the interpolator works as expected"""
        grid = mist.MISTBolometricCorrectionGrid(bands=("Gaia_G_EDR3",))
        
        # pick random points in the grid
        sample = grid.bc_grid.sample(10)
        teff, logg, feh, av = np.transpose(sample.index.tolist())
        expected = sample["Gaia_G_EDR3"].values

        self.assertTrue(np.allclose(
            grid.interp(teff, logg, feh, av)["Gaia_G_EDR3"].values,
            expected
        ))
        
        self.assertTrue(np.isclose(
            grid.interp(teff[0], logg[0], feh[0], av[0])["Gaia_G_EDR3"],
            expected[0]
        ))



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import tarfile
import logging

import requests
import pandas as pd
from scipy.interpolate import RegularGridInterpolator
import numpy as np


__all__ = ["MISTBolometricCorrectionGrid"]


# this is a mapping of photometric systems to the bands available in MIST BC tables
MIST_FILTER_SETS = {
    "UBVRIplus": [
        "Bessell_U", "Bessell_B", "Bessell_V", "Bessell_R", "Bessell_I", "2MASS_J", "2MASS_H", "2MASS_Ks",
        "Kepler_Kp", "Kepler_D51", "Hipparcos_Hp", "Tycho_B", "Tycho_V",
        "Gaia_G_DR2Rev", "Gaia_BP_DR2Rev", "Gaia_RP_DR2Rev", "Gaia_G_MAW", "Gaia_BP_MAWf", "Gaia_BP_MAWb",
        "Gaia_RP_MAW", "TESS", "Gaia_G_EDR3", "Gaia_BP_EDR3", "Gaia_RP_EDR3"
    ],
    "WISE":["WISE_W1", "WISE_W2", "WISE_W3", "WISE_W4"],
    "CFHT":["CFHT_u", "CFHT_g", "CFHT_r", "CFHT_i_new", "CFHT_i_old", "CFHT_z"],
    "DECam":["DECam_u", "DECam_g", "DECam_r", "DECam_i", "DECam_z", "DECam_Y"],
    "GALEX":["GALEX_FUV", "GALEX_NUV"],
    "JWST":[
        "F070W", "F090W", "F115W", "F140M", "F150W2", "F150W", "F162M", "F164N", "F182M", "F187N",
        "F200W", "F210M", "F212N", "F250M", "F277W", "F300M", "F322W2", "F323N", "F335M", "F356W",
        "F360M", "F405N", "F410M", "F430M", "F444W", "F460M", "F466N", "F470N", "F480M",
    ],
    "LSST": ["LSST_u", "LSST_g", "LSST_r", "LSST_i", "LSST_z", "LSST_y"],
    "PanSTARRS": ["PS_g", "PS_r", "PS_i", "PS_z", "PS_y", "PS_w", "PS_open"],
    "SkyMapper": ["SkyMapper_u", "SkyMapper_v", "SkyMapper_g", "SkyMapper_r", "SkyMapper_i", "SkyMapper_z"],
    "SPITZER": ["IRAC_3.6", "IRAC_4.5", "IRAC_5.8", "IRAC_8.0"],
    "UKIDSS": ["UKIDSS_Z", "UKIDSS_Y", "UKIDSS_J", "UKIDSS_H", "UKIDSS_K"],
    "SDSSugriz": ["SDSS_u", "SDSS_g", "SDSS_r", "SDSS_i", "SDSS_z"],
    "HST_ACSWF": ["ACS_WFC_F435W", "ACS_WFC_F475W", "ACS_WFC_F502N", "ACS_WFC_F550M", "ACS_WFC_F555W",
                  "ACS_WFC_F606W", "ACS_WFC_F625W", "ACS_WFC_F658N", "ACS_WFC_F660N", "ACS_WFC_F775W",
                  "ACS_WFC_F814W", "ACS_WFC_F850LP", "ACS_WFC_F892N"],
    "HST_ACSHR": ["ACS_HRC_F220W", "ACS_HRC_F250W", "ACS_HRC_F330W", "ACS_HRC_F344N", "ACS_HRC_F435W",
                  "ACS_HRC_F475W", "ACS_HRC_F502N", "ACS_HRC_F550M", "ACS_HRC_F555W", "ACS_HRC_F606W",
                  "ACS_HRC_F625W", "ACS_HRC_F658N", "ACS_HRC_F660N", "ACS_HRC_F775W", "ACS_HRC_F814W",
                  "ACS_HRC_F850LP", "ACS_HRC_F892N"],
    "HST_WFC3": ["WFC3_UVIS_F200LP", "WFC3_UVIS_F218W", "WFC3_UVIS_F225W", "WFC3_UVIS_F275W",
                 "WFC3_UVIS_F280N", "WFC3_UVIS_F300X", "WFC3_UVIS_F336W", "WFC3_UVIS_F343N",
                 "WFC3_UVIS_F350LP", "WFC3_UVIS_F373N", "WFC3_UVIS_F390M", "WFC3_UVIS_F390W",
                 "WFC3_UVIS_F395N", "WFC3_UVIS_F410M", "WFC3_UVIS_F438W", "WFC3_UVIS_F467M",
                 "WFC3_UVIS_F469N", "WFC3_UVIS_F475W", "WFC3_UVIS_F475X", "WFC3_UVIS_F487N",
                 "WFC3_UVIS_F502N", "WFC3_UVIS_F547M", "WFC3_UVIS_F555W", "WFC3_UVIS_F600LP",
                 "WFC3_UVIS_F606W", "WFC3_UVIS_F621M", "WFC3_UVIS_F625W", "WFC3_UVIS_F631N",
                 "WFC3_UVIS_F645N", "WFC3_UVIS_F656N", "WFC3_UVIS_F657N", "WFC3_UVIS_F658N",
                 "WFC3_UVIS_F665N", "WFC3_UVIS_F673N", "WFC3_UVIS_F680N", "WFC3_UVIS_F689M",
                 "WFC3_UVIS_F763M", "WFC3_UVIS_F775W", "WFC3_UVIS_F814W", "WFC3_UVIS_F845M",
                 "WFC3_UVIS_F850LP", "WFC3_UVIS_F953N", "WFC3_IR_F098M", "WFC3_IR_F105W", "WFC3_IR_F110W",
                 "WFC3_IR_F125W", "WFC3vIR_F126N", "WFC3_IR_F127M", "WFC3_IR_F128N", "WFC3_IR_F130N",
                 "WFC3_IR_F132N", "WFC3_IR_F139M", "WFC3_IR_F140W", "WFC3_IR_F153M", "WFC3_IR_F160W",
                 "WFC3_IR_F164N", "WFC3_IR_F167N" ],
    "HST_WFPC2": ["WFPC2_F218W", "WFPC2_F255W", "WFPC2_F300W", "WFPC2_F336W", "WFPC2_F439W", "WFPC2_F450W",
                  "WFPC2_F555W", "WFPC2_F606W", "WFPC2_F622W", "WFPC2_F675W", "WFPC2_F791W", "WFPC2_F814W",
                  "WFPC2_F850LP" ]
}


@dataclass
class MISTBolometricCorrectionGrid:
    """
    Download, cache, and ingest MIST bolometric correction grids.

    Parameters
    ----------
    bands : tuple[str]
        tuple of photometric bands to include (e.g. ("LSST_u", "BP", "G", "RP"))
    cache_dir
        directory where tarballs, extracted files, and HDF5s are stored
        (default: ~/.MIST_bc_grids)
    """
    bands: tuple[str] = ("G", "BP", "RP")
    cache_dir: Path = Path("~/.MIST_bc_grids").expanduser()
    rebuild: bool = False

    def __post_init__(self) -> None:
        self.cache_dir = Path(self.cache_dir).expanduser()
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        # find all of the necessary filter sets
        needed_filter_sets = set()
        for band in self.bands:
            found_it = False
            for filter_set in MIST_FILTER_SETS:
                if band in MIST_FILTER_SETS[filter_set]:
                    needed_filter_sets.add(filter_set)
                    found_it = True
                    break
        
            if not found_it:
                raise KeyError(f"band '{band}' not found in any MIST filter set")

        self.needed_filter_sets = needed_filter_sets

        # load all necessary filter sets and concatenate, keeping only requested bands
        dfs = [self.load_hdf5(filter_set) for filter_set in self.needed_filter_sets]
        df_cols = ["Rv", *self.bands]
        bc_grid = pd.concat(dfs, axis=1, copy=False)[df_cols]
        self.bc_grid = bc_grid.loc[:, ~bc_grid.columns.duplicated()]

        self._build_interpolators()


    def download_filter_set(self, filter_set: str) -> Path:
        """
        Download the MIST BC tarball for a given filter set (e.g. 'LSST').
        """
        tarball_path = self.cache_dir / f"{filter_set}.txz"

        if tarball_path.exists() and not self.rebuild:
            return tarball_path

        url = f"https://waps.cfa.harvard.edu/MIST/BC_tables/{filter_set}.txz"

        with requests.get(url, stream=True, timeout=60) as r:
            r.raise_for_status()
            with open(tarball_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=1024 * 1024):
                    if chunk:
                        f.write(chunk)

        return tarball_path

    def extract_filter_set(self, filter_set: str) -> Path:
        """
        Extract a downloaded tarball into a subdirectory of cache_dir.
        """
        extract_dir = self.cache_dir / filter_set
        extract_dir.mkdir(parents=True, exist_ok=True)

        if not self.rebuild and any(extract_dir.iterdir()):
            return extract_dir

        tarball_path = self.download_filter_set(filter_set)

        with tarfile.open(tarball_path, mode="r:*") as tf:
            tf.extractall(path=extract_dir)

        return extract_dir

    def _iter_data_files(self, folder: Path):
        for p in folder.rglob("*"):
            if p.is_file() and not p.name.startswith("."):
                yield p

    def read_filter_set(self, filter_set: str) -> pd.DataFrame:
        """
        Read all BC files for a filter set and concatenate into one DataFrame.
        """
        extract_dir = self.extract_filter_set(filter_set)

        dfs: list[pd.DataFrame] = []

        for fp in sorted(self._iter_data_files(extract_dir)):
            # read the header row (row 5 from the file)
            with open(fp, "r") as f:
                for _ in range(5 + 1):
                    header_line = f.readline()
            names = [s.replace("[Fe/H]", "feh") for s in header_line.strip().lstrip("#").split()]
            df = pd.read_csv(
                fp,
                sep="\\s+",
                comment="#",
                header=None,
                engine="python",
                names=names,
            )
            dfs.append(df)

        if not dfs:     # pragma: no cover
            raise FileNotFoundError(f"no BC files found in {extract_dir}")

        df = pd.concat(dfs, copy=False)
        df.set_index(["Teff", "logg", "feh", "Av"], inplace=True)
        return df

    def build_hdf5(self, filter_set: str) -> Path:
        """
        Build (or rebuild) a single HDF5 file for a filter set.
        """
        h5_path = self.cache_dir / f"{filter_set}.h5"

        if h5_path.exists() and not self.rebuild:
            return h5_path

        df = self.read_filter_set(filter_set)
        df.to_hdf(
            h5_path,
            key="bc",
            mode="w",
        )

        return h5_path

    def load_hdf5(self, filter_set: str) -> pd.DataFrame:
        """
        Load a previously-built HDF5 BC grid.
        """
        h5_path = self.cache_dir / f"{filter_set}.h5"
        if not h5_path.exists() or self.rebuild:
            self.build_hdf5(filter_set)
        return pd.read_hdf(h5_path, key="bc")
    
    def _build_interpolators(self) -> None:
        df = self.bc_grid.sort_index()

        teff = np.asarray(df.index.get_level_values("Teff").unique(), dtype=float)
        logg = np.asarray(df.index.get_level_values("logg").unique(), dtype=float)
        feh = np.asarray(df.index.get_level_values("feh").unique(), dtype=float)
        av = np.asarray(df.index.get_level_values("Av").unique(), dtype=float)

        teff.sort()
        logg.sort()
        feh.sort()
        av.sort()

        full_index = pd.MultiIndex.from_product(
            [teff, logg, feh, av],
            names=["Teff", "logg", "feh", "Av"],
        )
        dense = df.reindex(full_index)

        self._grid_axes = (teff, logg, feh, av)
        self._interpolators: dict[str, RegularGridInterpolator] = {}

        n_teff, n_logg, n_feh, n_av = len(teff), len(logg), len(feh), len(av)

        for band in self.bands:
            values_1d = dense[band].to_numpy(dtype=float, copy=False)
            values_4d = values_1d.reshape(n_teff, n_logg, n_feh, n_av)

            self._interpolators[band] = RegularGridInterpolator(
                self._grid_axes,
                values_4d,
                method="linear",
            )


    def interp(
        self,
        teff: float | np.ndarray,
        logg: float | np.ndarray,
        feh: float | np.ndarray,
        av: float | np.ndarray,
        bands: tuple[str, ...] | None = None,
        silence_bounds_warning: bool = False,
    ) -> pd.Series | pd.DataFrame:
        """
        Interpolate BCs at (Teff, logg, feh, Av) in that order.

        Returns
        -------
        - Series if all inputs are scalar
        - DataFrame if any input is array-like (one row per broadcasted point)
        """
        use_bands = self.bands if bands is None else bands

        teff_a = np.asarray(teff, dtype=float)
        logg_a = np.asarray(logg, dtype=float)
        feh_a = np.asarray(feh, dtype=float)
        av_a = np.asarray(av, dtype=float)

        teff_b, logg_b, feh_b, av_b = np.broadcast_arrays(teff_a, logg_a, feh_a, av_a)
        n = teff_b.size

        # warn the user if any points are out of bounds
        teff_min, teff_max = self._grid_axes[0][0], self._grid_axes[0][-1]
        logg_min, logg_max = self._grid_axes[1][0], self._grid_axes[1][-1]
        feh_min, feh_max = self._grid_axes[2][0], self._grid_axes[2][-1]
        av_min, av_max = self._grid_axes[3][0], self._grid_axes[3][-1]

        if not silence_bounds_warning:
            for var, label, min_val, max_val in [
                (teff_b, "Teff", teff_min, teff_max),
                (logg_b, "logg", logg_min, logg_max),
                (feh_b, "feh", feh_min, feh_max),
                (av_b, "Av", av_min, av_max),
            ]:
                n_out_of_bounds = ((var < min_val) | (var > max_val)).sum()
                if n_out_of_bounds > 0:
                    logging.getLogger("cogsworth").warning(
                        f"cogsworth warning: {n_out_of_bounds} out of bounds points for {label} when "
                        f"interpolating MIST BCs (valid range: {min_val} to {max_val}). Clipping to bounds."
                    )

        teff_b = np.clip(teff_b, teff_min, teff_max)
        logg_b = np.clip(logg_b, logg_min, logg_max)
        feh_b = np.clip(feh_b, feh_min, feh_max)
        av_b = np.clip(av_b, av_min, av_max)
                    
        pts = np.column_stack([
            teff_b.reshape(n),
            logg_b.reshape(n),
            feh_b.reshape(n),
            av_b.reshape(n),
        ])

        out = {b: self._interpolators[b](pts) for b in use_bands}

        # scalar -> Series
        if teff_b.shape == () and logg_b.shape == () and feh_b.shape == () and av_b.shape == ():
            return pd.Series({b: float(out[b][0]) for b in use_bands})

        # vectorised -> DataFrame
        return pd.DataFrame(out)



"""This file is used to test the runtime scaling of cogsworth when using multiple cores"""

import time
import cogsworth
import argparse
import numpy as np
import os

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Test the runtime scaling of cogsworth")
    parser.add_argument("-i", "--input", type=str, default=None, help="Input file to load in")
    parser.add_argument("-n", "--nbin", type=int, default=10000,
                        help="Number of binaries to simulate")
    parser.add_argument("-p", "--processes", type=int, default=1,
                        help="Number of processes to use")
    parser.add_argument("-w", "--weak-scaling", action="store_true",
                        help="Use weak scaling", default=False)
    args = parser.parse_args()

    if args.input is None and os.path.exists(f"runtime_base_{args.nbin}.h5"):
        args.input = f"runtime_base_{args.nbin}.h5"

    if args.input is None:
        print("Creating a base population")
        p = cogsworth.pop.Population(args.nbin, use_default_BSE_settings=True)
        p.sample_initial_binaries()
        p.sample_initial_galaxy()
        p = p[:args.nbin]
        p.save(f"runtime_base_{args.nbin}.h5")
    else:
        # load in a population of pre-sampled binaries
        print("Loading in a base population")
        p = cogsworth.pop.load(args.input, parts=["initial_binaries", "initial_galaxy"])

    # if we're testing weak scaling, we need to copy the population for each process
    if args.weak_scaling:
        copies = [p[:] for _ in range(args.processes)]
        p = cogsworth.pop.concat(*copies)

    p.processes = args.processes
    start = time.time()
    p.perform_stellar_evolution()
    cosmic_done = time.time()
    p.perform_galactic_evolution(progress_bar=False)
    gala_done = time.time()

    cosmic_time = cosmic_done - start
    gala_time = gala_done - cosmic_done
    total_time = gala_done - start

    print(f"{len(p)} binaries, with {args.processes}, took {total_time:1.2f} seconds to run ({cosmic_time:1.2f} for cosmic, {gala_time:1.2f} for gala)")

    suffix = "_weak" if args.weak_scaling else ""

    p.save(f"runtime_results/runtime_{args.nbin}_{args.processes}_pop{suffix}.h5", overwrite=True)

    np.save(f"runtime_results/runtimes_{args.nbin}_{args.processes}{suffix}.npy", [cosmic_time, gala_time, total_time])



from gala.units import galactic
import gala.potential as gp
import logging

__all__ = ["get_snapshot_potential"]


def get_snapshot_potential(snap, components=[{"label": "star", "attr": "s", "r_s": 3},
                                             {"label": "dark matter", "attr": "dm", "r_s": 10},
                                             {"label": "gas", "attr": "g", "r_s": 3}],
                           out_path=None, nmax=15, lmax=5, verbose=False):
    r"""Compute the potential of a snapshot of a hydrodynamical zoom-in simulation

    Parameters
    ----------
    snap : :class:`pynbody.snapshot.SimSnap`
        The snapshot
    components : `list`, optional
        List of components to add to the potential, each component specifies the label to use, attribute in
        the pynbody snap and scale radius to use, by default includes stars, dark matter and gas with scale
        radii of 3, 10 and 3 kpc respectively
    out_path : `str`, optional
        Path to save the potential (should be a .yml file), by default None
    nmax : `int`, optional
        Maximum value of $n$ for the radial expansion, by default 15
    lmax : `int`, optional
        Maximum value of $\ell$ for the spherical harmonics, by default 5
    verbose : `bool`, optional
        Whether to report on progress, by default False

    Returns
    -------
    pot : :class:`gala.potential.potential.CompositePotential`
        Potential of the snapshot
    """
    # start a composite potential
    pot = gp.CompositePotential()

    # compute the potential for each component
    for comp in components:
        if verbose:
            logging.getLogger("cogsworth").info(f"Computing potential for {comp['label']}")
        subsnap = getattr(snap, comp["attr"])

        # compute the coefficients for the SCF potential
        Snlm, Tnlm = gp.scf.compute_coeffs_discrete(xyz=subsnap["pos"],
                                                    mass=subsnap["mass"] / subsnap["mass"].sum(),
                                                    nmax=nmax, lmax=lmax, r_s=comp["r_s"], skip_m=True)

        # add the SCF potential to the composite potential
        pot[comp["label"]] = gp.scf.SCFPotential(m=subsnap["mass"].sum().tolist(), r_s=comp["r_s"],
                                                 Snlm=Snlm, Tnlm=Tnlm, units=galactic)

    # save the potential to a file if requested
    if out_path is not None:
        pot.save(out_path)

    return pot



import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import astropy.units as u
import astropy.constants as const

from .utils import kstar_translator, evol_type_translator

__all__ = ["plot_cmd", "plot_cartoon_evolution", "plot_galactic_orbit", "plot_hrd"]

# some styles to be used in all cogsworth plots
fs = 24
style_params = {
    'font.family': 'serif',
    'text.usetex': False,
    'figure.figsize': (12, 8),
    'axes.titlesize': 24,
    'legend.title_fontsize': 18,
    'legend.fontsize': 16,
    'axes.labelsize': 24,
    'xtick.labelsize': 21,
    'ytick.labelsize': 21,
    'axes.linewidth': 1.1,
    'xtick.major.size': 7,
    'xtick.minor.size': 4,
    'ytick.major.size': 7,
    'ytick.minor.size': 4,
    'xtick.minor.visible': True,
    'ytick.minor.visible': True,
}

# define a decorator for plotting functions which applies the style params in a context environment so that
# the style is only applied to plots made by cogsworth and not to any other plots the user might make
def cogsworth_plot_style(func):
    def wrapper(*args, **kwargs):
        with plt.rc_context(style_params):
            return func(*args, **kwargs)
    return wrapper


@cogsworth_plot_style
def plot_cmd(pop, m_filter="Gaia_G_EDR3", c_filter_1="Gaia_BP_EDR3", c_filter_2="Gaia_RP_EDR3",
             fig=None, ax=None, show=True, **kwargs):
    """Plot a colour-magnitude diagram for a population.

    Parameters
    ----------
    pop : :class:`~cogsworth.pop.Population`
        Population to plot.
    m_filter : `str`, optional
        Filter for the magnitude, by default "G"
    c_filter_1 : `str`, optional
        Filter for the first colour, by default "BP"
    c_filter_2 : `str`, optional
        Filter for the second colour, by default "RP"
    fig : :class:`~matplotlib.figure.Figure`, optional
        Figure on which to plot, by default will create a new one
    ax : :class:`~matplotlib.axes.Axes`, optional
        Axes on which to plot, by default will create a new one
    show : bool, optional
        Whether to immediately show the plot, by default True

    Returns
    -------
    fig, ax : :class:`~matplotlib.figure.Figure`
        Figure and axis that contains the plot
    """
    if fig is None or ax is None:
        fig, ax = plt.subplots(figsize=(7, 10))

        # flip the y-axis
        ax.invert_yaxis()

        # add a grid and axis labels
        ax.set_axisbelow(True)
        ax.grid(zorder=-1, linestyle="dotted", color="lightgrey")
        ax.set(xlabel=f'{c_filter_1} - {c_filter_2}', ylabel=m_filter)

        # create a custom colourmap based on the stellar types lists
        labels = [kstar_translator[i]["short"] for i in range(1, 13)]
        bounds = np.arange(1, 14)
        colours = [kstar_translator[i]["colour"] for i in range(1, 13)]
        cmap = mpl.colors.ListedColormap(colours)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
        cbar = fig.colorbar(mpl.cm.ScalarMappable(cmap=cmap, norm=norm),
                            spacing='proportional', label='Stellar type', ax=ax, pad=0.0)
        cbar.set_ticks(bounds[:-1] + 0.5, labels=labels, fontsize=0.5*fs)
        cbar.ax.tick_params(size=0.0)

    if "s" not in kwargs:
        kwargs["s"] = 10

    # plot everything up to neutron stars
    for kstar in range(13):
        # loop over bound binaries, disrupted primaries and disrupted secondaries
        for dis, marker, suffix in zip([False, True, True], ["o", "^", "v"], ["1", "1", "2"]):
            # select stars with the right kstar (based on which is brighter)
            mask = np.where(pop.observables["secondary_brighter"], 
                            pop.final_bpp["kstar_2"] == kstar,
                            pop.final_bpp["kstar_1"] == kstar)

            # apply disruption mask
            mask = (mask & pop.disrupted) if dis else (mask & ~pop.disrupted)

            ax.scatter(pop.observables[f'{c_filter_1}_app_{suffix}'][mask]
                       - pop.observables[f'{c_filter_2}_app_{suffix}'][mask],
                       pop.observables[f'{m_filter}_abs_{suffix}'][mask],
                       color=kstar_translator[kstar]["colour"],
                       marker=marker, **kwargs)

    if show:
        plt.show()

    return fig, ax


def _use_white_text(rgba):
    """Determines whether to use white text on a given background color based on the RGBA value.

    Parameters
    ----------
    rgba : `tuple`
        RGBA value

    Returns
    -------
    flag : `bool`
        True if white text should be used
    """
    r, g, b, _ = rgba
    return (r * 0.299 + g * 0.587 + b * 0.114) < 186 / 255


def _supernova_marker(ax, x, y, s):
    """Add a supernova marker to an ax

    Parameters
    ----------
    ax : :class:`~matplotlib.pyplot.axis`
        Axis on which to add marker
    x : `float`
        x coordinate
    y : `float`
        y coordinate
    s : `float`
        Size scale
    """
    ax.scatter(x, y, marker=(15, 1, 0), s=s * 6, zorder=-1,
               facecolor="#ebd510", edgecolor="#ebb810", linewidth=2)
    ax.scatter(x, y, marker=(10, 1, 0), s=s * 4, zorder=-1,
               facecolor="orange", edgecolor="#eb7e10", linewidth=2)


def _rlof_path(centre, width, height, m=1.5, flip=False):
    """Draw a path in the shape of RLOF (teardrop)

    Parameters
    ----------
    centre : `tuple`
        Centre of the teardrop (x, y)
    width : `float`
        How wide to make
    height : `float`
        How tall to make it
    m : `float`, optional
        The 'pointy-ness' of the teardrop, by default 1.5
    flip : `bool`, optional
        Whether to flip the direction, by default False

    Returns
    -------
    x, y : :class:`~numpy.ndarray`
        Path of the teardrop
    """
    t = np.linspace(0, 2 * np.pi, 1000)
    x = 0.5 * width * np.cos(t) * (-1 if flip else 1) + centre[0]
    y = height * np.sin(t) * np.sin(0.5 * t)**(m) + centre[1]

    return x, y


@cogsworth_plot_style
def plot_cartoon_evolution(bpp, bin_num, label_type="long", plot_title="Cartoon Binary Evolution",
                           y_sep_mult=1.5, offset=0.2, s_base=1000,
                           time_fs_mult=1.0, mass_fs_mult=1.0, kstar_fs_mult=1.0,
                           porb_fs_mult=1.0, label_fs_mult=1.0,
                           fig=None, ax=None, show=True):    # pragma: no cover
    """Plot COSMIC bpp output as a cartoon evolution

    Parameters
    ----------
    bpp : `pandas.DataFrame`
        COSMIC bpp table
    bin_num : `int`
        Binary number of the binary to plot
    label_type : `str`, optional
        What sort of annotated labels to use ["short", "long", "sentence"], by default "long"
    plot_title : `str`, optional
        Title to use for the plot, use "" for no title, by default "Cartoon Binary Evolution"
    y_sep_mult : `float`, optional
        Multiplier to use for the y separation (larger=more spread out steps, longer figure)
    offset : `float`, optional
        Offset from the centre for each of the stars (larger=wider binaries)
    s_base : `float`, optional
        Base scatter point size for the stars
    time_fs_mult : `float`, optional
        Multiplier for the time annotation fontsize
    mass_fs_mult : `float`, optional
        Multiplier for the mass annotation fontsize
    kstar_fs_mult : `float`, optional
        Multiplier for the kstar annotation fontsize
    porb_fs_mult : `float`, optional
        Multiplier for the porb annotation fontsize
    label_fs_mult : `float`, optional
        Multiplier for the evolution label fontsize
    fig : :class:`~matplotlib.pyplot.figure`, optional
        Figure on which to plot, by default will create a new one
    ax : :class:`~matplotlib.pyplot.axis`, optional
        Axis on which to plot, by default will create a new one
    show : `bool`, optional
        Whether to immediately show the plot, by default True

    Returns
    -------
    fig, ax : :class:`~matplotlib.pyplot.figure`, :class:`~matplotlib.pyplot.axis`
        Figure and axis of the plot
    """
    # extract the pertinent information from the bpp table
    df = bpp.loc[bin_num][["tphys", "mass_1", "mass_2", "kstar_1", "kstar_2", "porb", "sep",
                           "evol_type", "RRLO_1", "RRLO_2"]]

    # add some offset kstar columns to tell what type a star *previously* was
    df[["prev_kstar_1", "prev_kstar_2", "prev_evol_type"]] = df.shift(1, fill_value=0)[["kstar_1", "kstar_2",
                                                                                        "evol_type"]]

    # delete rows where RLOF ends immediately after a CE ends
    df = df[~((df["evol_type"] == 4) & (df["prev_evol_type"] == 8))]

    # count the number of evolution steps and start figure with size based on that
    total = len(df)
    if fig is None or ax is None:
        fig, ax = plt.subplots(figsize=(12, total * y_sep_mult))

    # instantiate some flags to track state of binary
    i = 0
    disrupted = False
    common_envelope = False
    rlof = False
    contact = False

    min_log10_sep = np.log10(df["sep"][df["sep"] > 0.0].min())
    max_log10_sep = np.log10(df["sep"].max())

    # group timesteps and row indices by time
    times, row_inds = [], []
    prev_time, rows, j = -1.0, [], 0
    for _, row in df.iterrows():
        if row["tphys"].round(2) == prev_time:
            rows.append(j)
        else:
            if prev_time >= 0.0:
                times.append(prev_time)
                row_inds.append(rows)
            rows = [j]
            prev_time = row["tphys"].round(2)
        j += 1
    # append the last set of rows
    times.append(prev_time)
    row_inds.append(rows)

    # annotate the time on the left side of the binary
    for time, inds in zip(times, row_inds):
        ax.annotate(f'{time:1.2e} Myr' if time > 1e4 else f'{time:1.2f} Myr',
                    xy=(-offset - 0.3 - (0.12 if len(inds) > 1 else 0),
                        total - np.mean(inds) * y_sep_mult), ha="right", va="center",
                    fontsize=0.4*fs*time_fs_mult, fontweight="bold", zorder=-1,
                    bbox=dict(boxstyle="round,pad=0.2", fc="white", ec="white") if len(inds) > 1 else None)
        # if there's more than one ind then plot a bracketed line connecting them
        if len(inds) > 1:
            ax.annotate('', xy=(-offset - 0.35, total - np.mean(inds) * y_sep_mult),
                        xytext=(-offset - 0.4, total - np.mean(inds) * y_sep_mult),
                        ha='center', va='center',
                        # 2.5 is a magic number here to make the bracket the right size
                        arrowprops=dict(arrowstyle=f'-[, widthB={y_sep_mult * len(inds) * 2.5}, lengthB=1',
                                        lw=1.5, color='k'))

    sep_offset = 0.2

    # go through each row of the evolution
    for _, row in df.iterrows():
        # use the translators to convert evol_type and kstars
        et_ind, k1, k2, pk1, pk2 = int(row["evol_type"]), kstar_translator[int(row["kstar_1"])],\
            kstar_translator[int(row["kstar_2"])], kstar_translator[int(row["prev_kstar_1"])],\
            kstar_translator[int(row["prev_kstar_2"])]
        et = evol_type_translator[et_ind]

        if row["sep"] > 0.0:
            sep_modifier = (np.log10(row["sep"]) - min_log10_sep) / (max_log10_sep - min_log10_sep)
            off_s = sep_offset * sep_modifier
        else:
            sep_modifier, off_s = None, 0.0

        # set disrupted, rlof and common-envelope flags are necessary
        if et_ind == 11 or row["porb"] < 0.0:
            disrupted = True
        if et_ind == 3 and (((row["RRLO_1"] >= 1.0) & (row["kstar_1"] < 13)) | (row["kstar_2"] < 13)):
            rlof = True
        if et_ind == 4:
            rlof = False
            contact = False
        if et_ind == 5:
            rlof = False
            contact = True
        if et_ind == 7:
            common_envelope = True
        if et_ind == 8:
            common_envelope = False
            rlof = False

        if row["RRLO_1"] < 1.0 and row["RRLO_2"] < 1.0:
            rlof = False
            common_envelope = False
            contact = False

        # check if either star is now a massless remnant
        mr_1 = k1["short"] == "MR"
        mr_2 = k2["short"] == "MR"
        ks_fontsize = 0.3 * fs * kstar_fs_mult

        # start an evolution label variable
        evol_label = et[label_type]

        # if the star just evolved then edit to label to explain what happened
        if et_ind == 2:
            which_star = "Primary" if k1 != pk1 else "Secondary"
            to_type = k1[label_type] if k1 != pk1 else k2[label_type]
            evol_label = f'{which_star} evolved to\n{to_type}'

        # annotate the evolution label on the right side of the binary
        ax.annotate(evol_label, xy=(0.5, total - i), va="center", fontsize=0.4*fs*label_fs_mult)

        # if we've got a common envelope then draw an ellipse behind the binary
        if common_envelope:
            envelope = mpl.patches.Ellipse(xy=(0, total - i),
                                           width=4 * offset + off_s, height=1.5 + off_s,
                                           facecolor="orange", edgecolor="none", zorder=-1, alpha=0.5)
            envelope_edge = mpl.patches.Ellipse(xy=(0, total - i),
                                                width=4 * offset + off_s, height=1.5 + off_s,
                                                facecolor="none", edgecolor="darkorange", lw=2)
            ax.add_artist(envelope)
            ax.add_artist(envelope_edge)

        # if either star is a massless remnant then we're just dealing with a single star now
        if mr_1 or mr_2:
            # plot the star centrally and a little larger
            ax.scatter(0, total - i, color=k1["colour"] if mr_2 else k2["colour"], s=s_base * 1.5)

            # label its stellar type if (a) it changed or (b) we're at the start/end of evolution
            if (k1 != pk1 and not mr_1) or (k2 != pk2 and not mr_2) or et_ind in [1, 10]:
                ax.annotate(k1["short"] if mr_2 else k2["short"], xy=(0, total - i),
                            ha="center", va="center", zorder=10, fontsize=ks_fontsize, fontweight="bold",
                            color="white" if _use_white_text(k1["colour"]
                                                             if mr_2 else k2["colour"]) else "black")

            # annotate the correct mass
            ax.annotate(f'{row["mass_1"] if mr_2 else row["mass_2"]:1.2f} ' + r'$\rm M_{\odot}$',
                        xy=(0, total - i - 0.45), ha="center", va="top", fontsize=0.3*fs*mass_fs_mult,
                        bbox=dict(boxstyle="round,pad=0.2", fc="white", ec="none", alpha=0.7)
                        if et_ind in [15, 16] else None)

            # if a supernova just happened then add an explosion marker behind the star
            if et_ind in [15, 16]:
                _supernova_marker(ax, 0, total - i, s_base)

        # otherwise we've got two stars
        else:
            contact_adjust = 0.25 if contact else 1.0

            # plot stars offset from the centre
            ax.scatter(0 - (offset + off_s) * contact_adjust, total - i,
                       color=k1["colour"], s=s_base, zorder=10)
            ax.scatter(0 + (offset + off_s) * contact_adjust, total - i,
                       color=k2["colour"], s=s_base, zorder=10)

            # annotate the mass (with some extra padding if there's RLOF)
            mass_y_offset = 0.35 if not (rlof and not common_envelope) else 0.5
            ax.annotate(f'{row["mass_1"]:1.2f} ' + r'$\rm M_{\odot}$',
                        xy=(0 - offset * contact_adjust - off_s, total - i - mass_y_offset),
                        ha="left" if common_envelope else "center", va="top", fontsize=0.3*fs*mass_fs_mult,
                        rotation=45 if contact else 0,
                        bbox=dict(boxstyle="round,pad=0.2", fc="white", ec="none", alpha=0.7)
                        if et_ind in [15, 16] else None)
            ax.annotate(f'{row["mass_2"]:1.2f} ' + r'$\rm M_{\odot}$',
                        xy=(0 + offset * contact_adjust + off_s, total - i - mass_y_offset),
                        ha="right" if common_envelope else "center", va="top", fontsize=0.3*fs*mass_fs_mult,
                        zorder=1000,
                        rotation=45 if contact else 0,
                        bbox=dict(boxstyle="round,pad=0.2", fc="white", ec="none", alpha=0.7)
                        if et_ind in [15, 16] else None)

            # if the primary type changed or we're at the start/end then label it
            if k1 != pk1 or et_ind in [1, 10]:
                ax.annotate(k1["short"], xy=(0 - offset * contact_adjust - off_s, total - i),
                            ha="center", va="center",
                            color="white" if _use_white_text(k1["colour"]) else "black",
                            zorder=10, fontsize=ks_fontsize, fontweight="bold")

            # if the secondary type changed or we're at the start/end then label it
            if k2 != pk2 or et_ind in [1, 10]:
                ax.annotate(k2["short"], xy=(0 + offset * contact_adjust + off_s, total - i),
                            ha="center", va="center",
                            color="white" if _use_white_text(k2["colour"]) else "black",
                            zorder=10, fontsize=ks_fontsize, fontweight="bold")

            # for bound binaries plot a line connecting them
            if not disrupted or et_ind == 11:
                ax.plot([0 - offset * contact_adjust - off_s, 0 + offset * contact_adjust + off_s], [total - i, total - i],
                        linestyle="--", zorder=-1, color="black")

            if et_ind == 11:
                ax.scatter(0, total - i, marker=(10, 1, 360 / 10 / 2), s=s_base / 2, zorder=-1,
                           facecolor="orange", edgecolor="none", linewidth=1)

            if not disrupted:
                # annotate the line with period, offset to one side if there's RLOF
                x = 0 if not (rlof and not common_envelope) else (-offset * contact_adjust / 4 if row["RRLO_1"] >= 1.0 else offset * contact_adjust / 4)
                p_lab = f'{row["porb"]:1.2e} days' if row["porb"] > 10000 or row["porb"] < 1\
                    else f'{row["porb"]:1.0f} days'
                ax.annotate(p_lab, xy=(x, total - i + 0.05), ha="center", va="bottom",
                            fontsize=0.2*fs*porb_fs_mult if row["porb"] > 10000 or row["porb"] < 1 else 0.3*fs*porb_fs_mult)

            # for non-common-envelope RLOF, plot a RLOF teardrop in the background
            if rlof and not common_envelope:
                # flip the shape depending on the direction
                if row["RRLO_1"] >= 1.0:
                    x, y = _rlof_path((0 - offset / 2.6, total - i), 2 * (offset + off_s),
                                      0.6 * (1 + off_s), flip=False)
                else:
                    x, y = _rlof_path((0 + offset / 2.6, total - i), 2 * (offset + off_s),
                                      0.6 * (1 + off_s), flip=True)
                ax.plot(x, y, color="darkorange", lw=2)
                ax.fill_between(x, y, color="orange", alpha=0.5, edgecolor="none", zorder=-2)

            # add supernova explosion markers as necessary
            if et_ind == 15:
                _supernova_marker(ax, 0 - offset * contact_adjust - off_s, total - i, s_base / 1.5)
            if et_ind == 16:
                _supernova_marker(ax, 0 + offset * contact_adjust + off_s, total - i, s_base / 1.5)

        # increment by multiplier
        i += y_sep_mult

    # clear off any x-ticks and axes
    ax.set_xlim(-1.5, 1.5)
    ax.set_xticks([])
    ax.axis("off")

    # annotate a title as the top
    ax.annotate(plot_title, xy=(0, total + 0.75), ha="center", va="center", fontsize=fs * 1.2)

    if show:
        plt.show()

    return fig, ax


@cogsworth_plot_style
def plot_galactic_orbit(primary_orbit, secondary_orbit=None,
                        t_min=0 * u.Myr, t_max=np.inf * u.Myr, show_start=True,
                        primary_kwargs={}, secondary_kwargs={}, start_kwargs={},
                        show_legend=True,
                        fig=None, axes=None, show=True):
    """Plot the galactic orbit of a binary system. This provides a wrapper around the gala
    :class:`~gala.dynamics.Orbit` method :meth:`~gala.dynamics.Orbit.plot`.

    Parameters
    ----------
    primary_orbit : :class:`~gala.dynamics.Orbit`
        Orbit of the primary star (or bound binary)
    secondary_orbit : :class:`~gala.dynamics.Orbit`, optional
        Orbit of the secondary star, by default None (no disruption)
    t_min : :class:`~astropy.units.Quantity` [time], optional
        Minimum time since the start of the orbits to plot, by default 0*u.Myr
    t_max : :class:`~astropy.units.Quantity` [time], optional
        Maximum time since the start of the orbits to plot, by default np.inf*u
    show_start : `bool`, optional
        Whether to plot a marker at the start of the orbits, by default True
    primary_kwargs : `dict`, optional
        Keyword arguments to pass to the primary orbit plot, by default {}
    secondary_kwargs : `dict`, optional
        Keyword arguments to pass to the secondary orbit plot, by default {}
    start_kwargs : `dict`, optional
        Keyword arguments to pass to the start marker plot, by default {}
    fig : :class:`~matplotlib.figure.Figure`, optional
        Figure on which to plot, by default will create a new one
    axes : :class:`~matplotlib.axes.Axes`, optional
        Axes on which to plot, by default will create a new one
    show : `bool`, optional
        Whether to immediately show the plot, by default

    Returns
    -------
    fig, axes : :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot
    """
    # create a mask for the times
    time_mask = ((primary_orbit.t - primary_orbit.t[0] >= t_min)
                 & (primary_orbit.t - primary_orbit.t[0] < t_max))

    if not time_mask.any():
        raise ValueError("No times in the specified range. Please check t_min and t_max.")

    # add a label to the primary orbit if not already present
    if "label" not in primary_kwargs:
        primary_kwargs["label"] = "Primary orbit"

    # plot the primary orbit and save the resulting figure
    fig = primary_orbit[time_mask].plot(axes=axes, **primary_kwargs)

    # same thing for secondary orbit (if it exists)
    if secondary_orbit is not None:
        if "label" not in secondary_kwargs:
            secondary_kwargs["label"] = "Secondary orbit"
        # find the first time at which the secondary orbit differs from the primary orbit
        t_diffs = np.where(np.any(secondary_orbit[time_mask].xyz != primary_orbit[time_mask].xyz, axis=0))[0]

        if len(t_diffs) != 0:
            t_diff = max(0, t_diffs[0] - 1)

            # only plot the secondary orbit if it differs from the primary orbit
            secondary_orbit[time_mask][t_diff:].plot(axes=fig.axes, **secondary_kwargs)

    # if we're showing the start position then plot a marker there
    if show_start:
        # use some default settings that the user can update
        full_start_kwargs = {
            "marker": 'o',
            "color": 'black',
            "s": 50,
            "label": "Start position"
        }
        full_start_kwargs.update(start_kwargs)

        primary_orbit[0].plot(axes=fig.axes, **full_start_kwargs)

    if show_legend:     # pragma: no cover
        handles, labels = axes[-1].get_legend_handles_labels()
        fig.legend(handles, labels, loc='upper center', ncol=5, fontsize=16)
        fig.subplots_adjust(top=0.9)

    # show the plot if desired
    if show:
        plt.show()

    return fig, fig.axes


def _LBV_limit(T_eff):
    """Compute the luminosity of the LBV limit given a certain temperature
    
    Parameters
    ----------
    T_eff : :class:`~astropy.units.Quantity` [temperature]
        Effective temperature(s) at which to compute the LBV limit

    Returns
    -------
    L_LBV : :class:`~astropy.units.Quantity` [luminosity]
        Luminosity of the LBV limit at the given temperature(s)
    """
    return np.maximum(
        np.sqrt(1e10 * u.Rsun**2 * u.Lsun * (4 * np.pi * const.sigma_sb * (T_eff * u.K)**4)).to(u.Lsun),
        np.repeat(6e5, len(T_eff)) * u.Lsun
    )


@cogsworth_plot_style
def plot_hrd(bcm, bin_num, show_primary=True, show_secondary=True,
             max_kstar=10, shade_LBV_regime=False, fig=None, ax=None, show=True, **ax_settings):
    """Plot the HR diagram of a binary evolution

    Parameters
    ----------
    bcm : `pandas.DataFrame`
        COSMIC bcm table
    bin_num : `int`
        Binary number of the binary to plot
    show_primary : `bool`, optional
        Whether to show the primary star, by default True
    show_secondary : `bool`, optional
        Whether to show the secondary star, by default True
    max_kstar : `int`, optional
        Maximum kstar value to plot, by default 10
    shade_LBV_regime : `bool`, optional
        Whether to shade the LBV regime, by default False
    fig : :class:`~matplotlib.figure.Figure`, optional
        Figure on which to plot, by default will create a new one
    ax : :class:`~matplotlib.axes.Axes`, optional
        Axis on which to plot, by default will create a new one
    show : `bool`, optional
        Whether to immediately show the plot, by default True
    **ax_settings : dict
        Additional keyword arguments for the axis settings (e.g. limits, ticks, etc.)

    Returns
    -------
    fig, ax : :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axis of the plot
    """

    if fig is None or ax is None:
        fig, ax = plt.subplots(figsize=(10, 16))

    kstars = set()
    min_log_teff, max_log_teff = np.inf, -np.inf

    if isinstance(bin_num, (int, np.int64)):
        bin_num = [bin_num]
    
    for bn in bin_num:
        for k, show_k in [(1, show_primary), (2, show_secondary)]:
            if not show_k:
                continue

            log_teff = np.log10(bcm[bcm["bin_num"] == bn][f"teff_{k}"].values)
            log_lum = np.log10(bcm[bcm["bin_num"] == bn][f"lum_{k}"].values)
            kstar = bcm[bcm["bin_num"] == bn][f"kstar_{k}"].values

            mask = kstar < max_kstar

            ax.scatter(log_teff[mask], log_lum[mask], c=[kstar_translator[k]["colour"] for k in kstar[mask]], s=5)
            ax.plot(log_teff[mask], log_lum[mask], color='grey', alpha=0.5, lw=0.5, zorder=-1)
            
            # add a marker to the start and end of the track
            ax.scatter(log_teff[mask][0], log_lum[mask][0], color="k" if k == 1 else "grey", marker="o", s=50)
            ax.scatter(log_teff[mask][-1], log_lum[mask][-1], color="k" if k == 1 else "grey", marker="o", s=50)

            kstars.update(set(kstar[mask]))
            min_log_teff = min(min_log_teff, log_teff[mask].min())
            max_log_teff = max(max_log_teff, log_teff[mask].max())

    # annotate the various kstar types
    for k in kstars:
        if k < max_kstar:
            ax.scatter([], [], color=kstar_translator[k]["colour"], label=kstar_translator[k]["short"], s=50)

    if show_primary:
        ax.scatter([], [], color="k", label=f"Start/end of primary", s=50)
    if show_secondary:
        ax.scatter([], [], color="grey", label=f"Start/end of secondary", s=50)
    ax.legend()

    ax.set(
        xlabel=r"$\log_{10}(T_{\rm eff} \, [\rm K])$",
        ylabel=r"$\log_{10}(L \, [\rm L_{\odot}])$",
        **ax_settings
    )

    if shade_LBV_regime:
        ylims = ax.get_ylim()

        # make some fairly reasonable temperature range
        T_eff_range = np.logspace(min_log_teff - 0.1, max_log_teff + 0.1, 1000)
        LBV_log_L = np.log10(_LBV_limit(T_eff_range).value)

        if ylims[0] < LBV_log_L.min() < ylims[1]:
            # plot a line at the HD limit and fill the area
            ax.plot(np.log10(T_eff_range), LBV_log_L, color="grey", linestyle="dotted")
            ax.fill_between(np.log10(T_eff_range), LBV_log_L, 9, color="black", lw=2, alpha=0.03)

            # annotate the regime with a custom location
            ax.annotate("LBV Regime", xy=(0.77, 0.93), xycoords="axes fraction", ha="center", va="center",
                        fontsize=20, color="grey", zorder=10)
            ax.set_ylim(ylims)

    
    # flip x-axis if necessary
    if ax.get_xlim()[0] < ax.get_xlim()[1]:
        ax.invert_xaxis()

    if show:        # pragma: no cover
        plt.show()
    return fig, ax



import numpy as np
import pandas as pd
import astropy.units as u
import astropy.constants as const
from astropy.coordinates import SkyCoord
from copy import copy

import logging
from cogsworth.tests.optional_deps import check_dependencies
from cogsworth.obs.mist import MISTBolometricCorrectionGrid

import sys
import os

__all__ = ["get_log_g", "get_absolute_bol_mag", "get_apparent_mag", "get_absolute_mag", "add_mags",
           "get_extinction", "get_photometry"]


def get_log_g(mass, radius):
    """Computes log of the surface gravity in cgs

    Parameters
    ----------
    mass : :class:`~astropy.units.Quantity` [mass]
        Mass of the star
    radius : :class:`~astropy.units.Quantity` [radius]
        Radius of the star

    Returns
    -------
    log g : :class:`~numpy.ndarray`
        Log of the surface gravity in cgs
    """
    g = const.G * mass / radius**2

    # avoid division by zero errors (for massless remnants)
    with np.errstate(divide='ignore'):
        return np.log10(g.cgs.value)


def get_absolute_bol_mag(lum):
    """Computes the absolute bolometric magnitude following
    `IAU Resolution B2 <https://www.iau.org/news/announcements/detail/ann15023/>`_

    Parameters
    ----------
    lum : :class:`~astropy.units.Quantity` [luminosity]
        Luminosity of the star

    Returns
    -------
    M_bol : :class:`~numpy.ndarray`
        Absolute bolometric magnitude
    """
    zero_point_lum = 3.0128e28 * u.watt
    return -2.5 * np.log10(lum / zero_point_lum)


def get_apparent_mag(M_abs, distance):
    """Convert absolute magnitude to apparent magnitude

    Parameters
    ----------
    M_abs : :class:`~numpy.ndarray`
        Absolute magnitude
    distance : :class:`~astropy.units.Quantity` [length]
        Distance

    Returns
    -------
    m_app : :class:`~numpy.ndarray`
        Apparent magnitude
    """
    finite_distance = np.isfinite(distance)
    m_app = np.repeat(np.inf, len(distance))
    m_app[finite_distance] = M_abs[finite_distance] + 5 * np.log10(distance[finite_distance] / (10 * u.pc))
    return m_app


def get_absolute_mag(m_app, distance):
    """Convert apparent magnitude to absolute magnitude

    Parameters
    ----------
    M_abs : :class:`~numpy.ndarray`
        Apparent magnitude
    distance : :class:`~astropy.units.Quantity` [length]
        Distance

    Returns
    -------
    m_app : :class:`~numpy.ndarray`
        Absolute magnitude
    """
    M_abs = m_app - 5 * np.log10(distance / (10 * u.pc))
    return M_abs


def add_mags(*mags, remove_nans=True):
    """Add any number of magnitudes

    Parameters
    ----------
    *mags : `list` or `np.array` or `float` or `int`
        A series of magnitudes. If arrays are given then all must have the same length. If a mixture of single
        values and arrays are given then the single values will be added to each array element
    remove_nans : `bool`, optional
        Whether to remove NaNs from the total (if not the total will be NaN), by default True

    Returns
    -------
    total_mag : :class:`~numpy.ndarray`
        Total magnitude

    Raises
    ------
    ValueError
        If any magnitude is an invalid type
    AssertionError
        If any magnitude array has a different length from another
    """
    total_mag = 0

    # convert lists to numpy arrays
    mags = list(mags)
    for i in range(len(mags)):
        if isinstance(mags[i], list):
            mags[i] = np.array(mags[i])
        if isinstance(mags[i], int):
            mags[i] = float(mags[i])

    # check for dodgy input
    if isinstance(mags[0], (list, np.ndarray)):
        length = len(mags[0])
        for mag in mags[1:]:
            assert len(mag) == length, ("All magnitude arrays must have the same length - one array is of "
                                        f"length {length} but another is {len(mag)}")

    # go through each supplied magnitude
    for mag in mags:
        if not isinstance(mag, (np.ndarray, float, int)):
            raise ValueError(("All `mag`s must either be a list, numpy array, float or an int. Unfortunately "
                              f"for us both, what you have given me is of type `{type(mag).__name__}`..."))

        # compute the default additional magnitude
        additional = 10**(-mag * 0.4)

        # if you want to remove any NaNs then do so
        if remove_nans:
            if isinstance(mag, np.ndarray):
                additional[np.isnan(mag)] = 0.0
            elif np.isnan(mag):
                additional = 0.0
        total_mag += additional

    # hide divide by zero errors (since inf magnitude is correct when all mags are NaN)
    with np.errstate(divide='ignore'):
        return -2.5 * np.log10(total_mag)


def get_extinction(coords):     # pragma: no cover
    """Calculates the visual extinction values for a set of coordinates

    Reddening due to dust is calculated using the Bayestar dustmap. Then the conversion from this to a visual
    extension is done assuming a total-to-selective extinction ratio of 3.3, as is used by
    `Green+2019 <https://iopscience.iop.org/article/10.3847/1538-4357/ab5362#apjab5362s2>`_

    .. warning::
        The dustmap used only covers declinations > -30 degrees, any supplied coordinates below this will be
        reflected around the galactic plane (any of [(-l, b), (l, -b), (-l, -b)]) and the dust at these
        locations will be used instead

    Parameters
    ----------
    coords : :class:`~astropy.coordinates.SkyCoord`
        The coordinates at which you wish to calculate extinction values

    Returns
    -------
    Av : :class:`~numpy.ndarray`
        Visual extinction values for each set of coordinates
    """
    assert check_dependencies("dustmaps")
    from dustmaps.bayestar import BayestarQuery

    # following section performs reflections for coordinates below -30 deg declination
    # convert to galactic coordinates
    galactic = coords.galactic

    # try all possible reflections about the galactic plane
    for ref_l, ref_b in [(-1, 1), (1, -1), (-1, -1)]:
        # check which things are too low for the dustmap
        too_low = galactic.icrs.dec < (-30 * u.deg)

        # if everything is fine now then stop
        if not too_low.any():
            break

        # apply the reflection to the subset that are too low
        reflected = galactic[too_low]
        reflected.data.lon[()] *= ref_l
        reflected.data.lat[()] *= ref_b
        reflected.cache.clear()

        # check which are fixed now, and reverse the reflection if not
        fixed = reflected.icrs.dec > (-30 * u.deg)
        reflected.data.lon[~fixed] *= ref_l
        reflected.data.lat[~fixed] *= ref_b

        # set the data back in the main coord object
        galactic.data.lon[too_low] = reflected.data.lon
        galactic.data.lat[too_low] = reflected.data.lat

        # clear the cache to ensure consistency
        galactic.cache.clear()

    # briefly disable print statements to hide the messy output from dustmaps
    sys.stdout = open(os.devnull, 'w')

    bayestar = BayestarQuery(max_samples=2, version='bayestar2019')

    # calculate the reddening due to dust
    ebv = bayestar(galactic, mode='random_sample')

    # bring print statements back
    sys.stdout = sys.__stdout__

    # convert this to a visual extinction
    Av = 3.3 * ebv
    return Av


def get_photometry(filters, population=None, final_bpp=None, final_pos=None, distances=None,
                   ignore_extinction=False, assume_mw_galactocentric=False,
                   main_filter=None, silence_bounds_warning=False):
    """Computes photometry subject to dust extinction using the MIST boloemtric correction grid

    Parameters
    ----------
    filters : `list` of `str`
        Which filters to compute photometry for (e.g. ['Gaia_G_EDR3', 'Gaia_BP_EDR3', 'Gaia_RP_EDR3'])
    population : :class:`~cogsworth.pop.Population`
        The population for which to compute photometry (either supply this or a final_bpp and final_pos)
    final_bpp : :class:`~pandas.DataFrame`
        A dataset of COSMIC binaries at present day - must include these columns: ["sep", "metallicity"] and
        for each star it must have the columns ["teff", "lum", "mass", "rad", "kstar"]
    final_pos : :class:`~astropy.quantity.Quantity`
        Final positions of each system in the galactocentric frame.
        The first `len(self)` entries of each are for bound binaries or primaries, then the final
        `self.disrupted.sum()` entries are for disrupted secondaries. Any missing orbits (where orbit=None
        will be set to `np.inf` for ease of masking.
    ignore_extinction : `bool`
        Whether to ignore extinction
    assume_mw_galactocentric : `bool`
        Whether to assume the population positions/distances are in the Milky Way galactocentric frame
    main_filter : `str`
        The main filter to use for calculating which star is observed as the brighter one
    silence_bounds_warning : `bool`
        Whether to silence the out-of-bounds warning from the MIST bolometric correction grid


    Returns
    -------
    photometry : :class:`~pandas.DataFrame`
        Photometry and extinction information for supplied COSMIC binaries in desired `filters`
    """
    # check that the input is valid
    if population is None and (final_bpp is None or final_pos is None):
        raise ValueError("Either a population or final_bpp and final_pos must be supplied")
    if distances is None and not assume_mw_galactocentric:
        raise ValueError("Must supply either distances or have `assume_mw_galactocentric=True`")
    if not ignore_extinction and not assume_mw_galactocentric:
        raise ValueError("Cannot calculate extinction due to dust without `assume_mw_galactocentric=True`")
    main_filter = filters[0] if main_filter is None else main_filter

    if population is not None:
        final_bpp = population.final_bpp
        final_pos = population.final_pos
        disrupted = population.disrupted
    else:
        disrupted = final_bpp["sep"].values < 0.0

    if assume_mw_galactocentric:
        final_coords = SkyCoord(x=final_pos[:, 0], y=final_pos[:, 1], z=final_pos[:, 2],
                                frame="galactocentric", unit=u.kpc, representation_type="cartesian")

    # set up empty photometry table
    photometry = pd.DataFrame()

    if not ignore_extinction:       # pragma: no cover
        # get extinction for bound binaries and primary of disrupted binaries
        photometry['Av_1'] = get_extinction(final_coords[:len(final_bpp)])

        # get extinction for secondaries of disrupted binaries (leave as np.inf otherwise)
        photometry['Av_2'] = np.repeat(np.inf, len(final_bpp))
        photometry.loc[disrupted, "Av_2"] = get_extinction(final_coords[len(final_bpp):])

        # ensure extinction remains in MIST grid range (<= 6) and is not NaN
        photometry.loc[photometry.Av_1 > 6, ['Av_1']] = 6
        photometry.loc[photometry.Av_2 > 6, ['Av_2']] = 6
        photometry = photometry.fillna(6)
    else:
        photometry['Av_1'] = np.zeros(len(final_bpp))
        photometry['Av_2'] = np.zeros(len(final_bpp))

    # get Fe/H using e.g. Bertelli+1994 Eq. 10 (assuming all stars have the solar abundance pattern)
    Z_sun = 0.0142
    FeH = np.log10(final_bpp["metallicity"].values / Z_sun)

    # set up MIST bolometric correction grid
    bc_grid = MISTBolometricCorrectionGrid(bands=filters)
    bc = {
        "app": [None, None],
        "abs": [None, None]
    }

    if distances is None:
        distances = final_coords.icrs.distance
    primary_distances = distances[:len(final_bpp)]
    # secondary distances are the same whilst bound but different when disrupted
    secondary_distances = copy(primary_distances)
    secondary_distances[disrupted] = distances[len(final_bpp):]

    # for each star in the (possibly disrupted/merged) binary
    for ind, dist in zip([1, 2], [primary_distances, secondary_distances]):
        # calculate the surface gravity if necessary
        if f"log_g_{ind}" not in final_bpp:
            final_bpp.insert(len(final_bpp.columns), f"log_g_{ind}",
                             get_log_g(mass=final_bpp[f"mass_{ind}"].values * u.Msun,
                                       radius=final_bpp[f"rad_{ind}"].values * u.Rsun))

        # get the bolometric corrections from MIST isochrones
        bc["app"][ind - 1] = bc_grid.interp(teff=final_bpp[f"teff_{ind}"].values,
                                            logg=final_bpp[f"log_g_{ind}"].values,
                                            feh=FeH, av=photometry[f"Av_{ind}"], bands=filters,
                                            silence_bounds_warning=silence_bounds_warning).values
        bc["abs"][ind - 1] = bc_grid.interp(teff=final_bpp[f"teff_{ind}"].values,
                                            logg=final_bpp[f"log_g_{ind}"].values,
                                            feh=FeH, av=np.zeros(len(final_bpp)), bands=filters,
                                            silence_bounds_warning=silence_bounds_warning).values

        # calculate the absolute bolometric magnitude and set any BH or massless remnants to invisible
        photometry[f"M_abs_{ind}"] = get_absolute_bol_mag(lum=final_bpp[f"lum_{ind}"].values * u.Lsun)
        photometry.loc[np.isin(final_bpp[f"kstar_{ind}"].values, [13, 14, 15]), f"M_abs_{ind}"] = np.inf

        # convert the absolute magnitude to an apparent magnitude
        photometry[f"m_app_{ind}"] = get_apparent_mag(M_abs=photometry[f"M_abs_{ind}"].values,
                                                      distance=dist)

    # go through each filter
    for i, filter in enumerate(filters):
        for prefix, mag_type in [("m", "app"), ("M", "abs")]:
            # apply the bolometric corrections to the apparent magnitude of each star
            filter_mags = [photometry[f"{prefix}_{mag_type}_{ind}"].values - bc[mag_type][ind - 1][:, i]
                           for ind in [1, 2]]

            # total the magnitudes (removing any NaNs)
            total_filter_mag = add_mags(*filter_mags, remove_nans=True)

            # default to assuming all systems are bound - in this case total magnitude is listed
            # in primary filter mag, and secondary is non-existent
            photometry[f"{filter}_{mag_type}_1"] = total_filter_mag
            photometry[f"{filter}_{mag_type}_2"] = np.repeat(np.inf, len(photometry))

            # for disrupted systems, change filter apparent magnitudes to the values for each individual star
            for ind in [1, 2]:
                photometry.loc[disrupted, f"{filter}_{mag_type}_{ind}"] = filter_mags[ind - 1][disrupted]

            # for the G filter in particular, see which temperature/log-g is getting measured
            if filter == main_filter and mag_type == "app":
                # by default assume the primary is dominant
                photometry["teff_obs"] = final_bpp["teff_1"].values
                photometry["log_g_obs"] = final_bpp["log_g_1"].values

                # overwrite any values where the secondary is brighter
                two_is_brighter = (filter_mags[1] < filter_mags[0]) | (np.isnan(filter_mags[0])
                                                                       & ~np.isnan(filter_mags[1]))
                photometry["secondary_brighter"] = two_is_brighter
                photometry.loc[two_is_brighter, "teff_obs"] = final_bpp["teff_2"].values[two_is_brighter]
                photometry.loc[two_is_brighter, "log_g_obs"] = final_bpp["log_g_2"].values[two_is_brighter]

    return photometry



import numpy as np
import astropy.units as u
import pandas as pd
import gala.dynamics as gd
import gala.integrate as gi

from multiprocessing import Pool
from tqdm import tqdm
import warnings
import logging

__all__ = ["rewind_to_formation"]


# define a function to integrate the orbit of a particle backwards in time
def _tohspans(pot, wf, t1, t2, dt):
    """tohspans = snapshot backwards in time lol"""
    return pot.integrate_orbit(wf, t1=t1, t2=t2, dt=dt,
                               Integrator=gi.DOPRI853Integrator, store_all=False)[-1]


def rewind_to_formation(subsnap, pot, dt=-1 * u.Myr, processes=1):
    """Rewind a snapshot to the time of formation of each particle

    Parameters
    ----------
    subsnap : :class:`pynbody.snapshot.SimSnap`
        A subset of a snapshot containing only the particles to rewind
    pot : :class:`gala.potential.potential.CompositePotential`
        The potential of the snapshot
    dt : :class:`~astropy.units.Quantity` [time], optional
        Timestep size, by default -1*u.Myr
    processes : `int`, optional
        How many processes to use, by default 1 (run single-threaded)

    Returns
    -------
    init_particles : :class:`~pandas.DataFrame`
        A dataframe containing the initial mass, metallicity, formation time, ID,
        position and velocity of each particle
    """
    # get the final time of the snapshot in Myr
    final_time = subsnap.properties["time"].in_units("Myr") * u.Myr

    # convert the final particles to a phase-space position
    wf = gd.PhaseSpacePosition(pos=np.transpose(subsnap["pos"].in_units("kpc").tolist()) * u.kpc,
                               vel=np.transpose(subsnap["vel"].in_units("km s**-1").tolist()) * u.km / u.s)
    # store their formation times in Myr
    tforms = subsnap["tform"].in_units("Myr").tolist() * u.Myr

    def args(wf, tforms):       # pragma: no cover
        for w, t in zip(wf, tforms):
            yield pot, w, final_time, t, dt

    # integrate the orbits of the particles backwards in time to their formation times
    # if the user wants to use multiple processes, do so
    if processes > 1:
        with Pool(processes) as pool:
            w0 = list(pool.starmap(_tohspans, tqdm(args(wf, tforms), total=len(subsnap))))

    # otherwise, just do it single-threaded
    else:
        w0 = [_tohspans(pot, wf[i], final_time, tforms[i], dt) for i in tqdm(range(len(subsnap)))]

    # check if the formation masses are available, warn if not
    if "massform" not in subsnap.all_keys():
        mass = subsnap["mass"].in_units("Msol").tolist()
        logging.getLogger("cogsworth").info("Formation masses (`massform`) not found, using present day masses instead")
    else:
        mass = subsnap["massform"].in_units("Msol").tolist()

    # create a df of the initial particles and return
    init_particles = np.zeros((len(subsnap), 10))
    init_particles[:, 0] = mass
    init_particles[:, 1] = subsnap["metals"] if np.ndim(subsnap["metals"]) == 1 else subsnap["metals"][:, 0]
    init_particles[:, 2] = subsnap["tform"].in_units("Gyr").tolist()
    init_particles[:, 3] = subsnap["iord"]

    for i in range(len(w0)):
        init_particles[i, 4:7] = w0[i].xyz.to(u.kpc).value
        init_particles[i, 7:10] = w0[i].vel.d_xyz.to(u.km / u.s).value

    df = pd.DataFrame(init_particles, columns=["mass", "Z", "t_form", "id",
                                               "x", "y", 'z', 'v_x', 'v_y', 'v_z'])
    # change ID column to integers and use as index
    df = df.astype({"id": int}, copy=False)
    df.set_index("id", inplace=True)
    return df



import numpy as np
import unittest
import cogsworth.sfh as sfh
import os
import astropy.units as u
import gala.potential as gp
from gala.units import galactic
import matplotlib.pyplot as plt
import tempfile

class Test(unittest.TestCase):
    def test_basic_class(self):
        """Check that base class can't be used alone"""

        g = sfh.StarFormationHistory(size=10000, components=[""], component_masses=[1],
                                     immediately_sample=False)
        it_broke = False
        try:
            g.draw_lookback_times()
        except NotImplementedError:
            it_broke = True
        self.assertTrue(it_broke)

        it_broke = False
        try:
            g.draw_radii()
        except NotImplementedError:
            it_broke = True
        self.assertTrue(it_broke)

        it_broke = False
        try:
            g.draw_phi()
        except NotImplementedError:
            it_broke = True
        self.assertTrue(it_broke)

        it_broke = False
        try:
            g.draw_heights()
        except NotImplementedError:
            it_broke = True
        self.assertTrue(it_broke)

        it_broke = False
        try:
            g.get_metallicity()
        except NotImplementedError:
            it_broke = True
        self.assertTrue(it_broke)

    def test_burst_uniform_disc(self):
        """Ensure the burst uniform disc class works"""
        g = sfh.BurstUniformDisc(size=10000,
                                 t_burst=5 * u.Gyr,
                                 R_max=20 * u.kpc,
                                 z_max=1 * u.kpc,
                                 Z_all=0.02)
        self.assertTrue(np.all(g.tau == 5 * u.Gyr))
        self.assertTrue(np.all(g.z <= 1 * u.kpc))
        self.assertTrue(np.all(g.rho <= 20 * u.kpc))
        self.assertTrue(np.all(g.Z == 0.02))

    def test_constant_uniform_disc(self):
        """Ensure the constant uniform disc class works"""
        g = sfh.ConstantUniformDisc(size=10000,
                                    t_burst=5 * u.Gyr,
                                    R_max=20 * u.kpc,
                                    z_max=1 * u.kpc,
                                    Z_all=0.02)
        self.assertTrue(np.all(g.tau <= 5 * u.Gyr))
        self.assertTrue(np.all(g.z <= 1 * u.kpc))
        self.assertTrue(np.all(g.rho <= 20 * u.kpc))
        self.assertTrue(np.all(g.Z == 0.02))

    def test_constant_plummer_sphere(self):
        """Ensure the constant Plummer sphere class works"""
        g = sfh.ConstantPlummerSphere(size=10000,
                                      tau_min=0 * u.Gyr,
                                      tau_max=10 * u.Gyr,
                                      Z_all=0.02,
                                      M=1e10*u.Msun,
                                      a=5.0*u.kpc,
                                      r_trunc=30 * u.kpc)
        self.assertTrue(np.all(g.tau >= 0 * u.Gyr))
        self.assertTrue(np.all(g.tau <= 10 * u.Gyr))
        self.assertTrue(np.all(g.Z == 0.02))
        self.assertTrue(np.all(np.sqrt(g.x**2 + g.y**2 + g.z**2) <= 30 * u.kpc))

    def test_plummer_indexing(self):
        """Ensure the constant Plummer sphere class works"""
        g = sfh.ConstantPlummerSphere(size=100,
                                      tau_min=0 * u.Gyr,
                                      tau_max=10 * u.Gyr,
                                      Z_all=0.02,
                                      M=1e10*u.Msun,
                                      a=5.0*u.kpc,
                                      r_trunc=30 * u.kpc)
        g_subset = g[:5]
        self.assertTrue(g_subset.size == 5)
        self.assertTrue(np.all(g.tau[:5] == g_subset.tau))

    def test_bad_inputs(self):
        """Ensure the classes fail with bad input"""
        g = sfh.Wagg2022(size=None, immediately_sample=False)
        it_broke = False
        try:
            g.sample()
        except ValueError:
            it_broke = True
        self.assertTrue(it_broke)

        g = sfh.Wagg2022(size=100, components=None, component_masses=None, immediately_sample=False)
        it_broke = False
        try:
            g.sample()
        except ValueError:
            it_broke = True
        self.assertTrue(it_broke)

    def test_valid_ranges(self):
        """Check that the drawn variables have valid ranges"""
        g = sfh.Wagg2022(size=10000)

        self.assertTrue((g.tau.min() >= 0) & (g.tau.max() <= g.galaxy_age))
        self.assertTrue(g.Z.min() >= 0.0)

    def test_io(self):
        """Check that a galaxy can be saved and re-loaded"""
        g = sfh.Wagg2022(size=10000)

        with tempfile.TemporaryDirectory() as tmpdir:
            g.save(os.path.join(tmpdir, "testing-galaxy-io"))
            g_loaded = sfh.load(os.path.join(tmpdir, "testing-galaxy-io"))

        self.assertTrue(np.all(g.tau == g_loaded.tau))
        self.assertTrue(np.all(g.rho == g_loaded.rho))
        self.assertTrue(np.all(g.z == g_loaded.z))

    def test_io_custom_sfh(self):
        """Check you can save a custom SFH class"""
        class Fixed_Z(sfh.Wagg2022):
            def get_metallicity(self):
                return 0.02

        g = Fixed_Z(size=10000)

        with tempfile.TemporaryDirectory() as tmpdir:
            g.save(os.path.join(tmpdir, "testing-galaxy-io-custom"))
            g_loaded = sfh.load(os.path.join(tmpdir, "testing-galaxy-io-custom"))

        self.assertTrue(np.all(g.tau == g_loaded.tau))
        self.assertTrue(np.all(g.rho == g_loaded.rho))
        self.assertTrue(np.all(g.z == g_loaded.z))

    def test_io_custom_sfh_with_params(self):
        """Check you can save a custom SFH class with its own parameters"""
        class Fixed_Z(sfh.Wagg2022):
            def __init__(self, fixed_Z, **kwargs):
                self.fixed_Z = fixed_Z
                super().__init__(**kwargs)

            def get_metallicity(self):
                return self.fixed_Z

        g = Fixed_Z(size=10000, fixed_Z=0.02)

        with tempfile.TemporaryDirectory() as tmpdir:
            g.save(os.path.join(tmpdir, "testing-galaxy-io-custom-params"))
            g_loaded = sfh.load(os.path.join(tmpdir, "testing-galaxy-io-custom-params"))    

        self.assertTrue(np.all(g.tau == g_loaded.tau))
        self.assertTrue(np.all(g.rho == g_loaded.rho))
        self.assertTrue(np.all(g.z == g_loaded.z))

    def test_io_SB15(self):
        """Check you can save and load the SB15 class"""
        g = sfh.SandersBinney2015(size=1000, time_bins=1, potential=gp.MilkyWayPotential(version='v2'))

        with tempfile.TemporaryDirectory() as tmpdir:
            g.save(os.path.join(tmpdir, "testing-galaxy-io-sb15"))
            g_loaded = sfh.load(os.path.join(tmpdir, "testing-galaxy-io-sb15"))

        self.assertTrue(np.all(g.tau == g_loaded.tau))
        self.assertTrue(np.all(g.rho == g_loaded.rho))
        self.assertTrue(np.all(g.z == g_loaded.z))
        self.assertTrue(g.potential == g_loaded.potential)

    def test_getters(self):
        """Test getting attributes"""
        it_broke = False
        try:
            g = sfh.Wagg2022(size=10, immediately_sample=False)
            len(g)
            g.components
            g.component_masses
            g.tau
            g = sfh.Wagg2022(size=10, immediately_sample=False)
            g.Z
            g = sfh.Wagg2022(size=10, immediately_sample=False)
            g.x
            g = sfh.Wagg2022(size=10, immediately_sample=False)
            g.y
            g = sfh.Wagg2022(size=10, immediately_sample=False)
            g.z
            g = sfh.Wagg2022(size=10, immediately_sample=False)
            g.positions
            g = sfh.Wagg2022(size=10, immediately_sample=False)
            g.which_comp
        except Exception as e:
            print(e)
            it_broke = True
        self.assertFalse(it_broke)

    def test_setters(self):
        """Test setting attributes"""
        g = sfh.Wagg2022(size=10000, immediately_sample=False)
        g.size = 100
        self.assertTrue(g.size == 100)

        # make sure it crashes for invalid inputs
        it_broke = False
        try:
            g.size = -1
        except ValueError:
            it_broke = True
        self.assertTrue(it_broke)

        it_broke = False
        try:
            g.size = "not an int"
        except ValueError:
            it_broke = True
        self.assertTrue(it_broke)

    def test_custom_galaxy(self):
        """Test saving a custom galaxy class"""

        class Custom(sfh.Wagg2022):
            def __init__(self, size, components=["low_alpha_disc"], component_masses=[1], **kwargs):
                super().__init__(size, components, component_masses, **kwargs)

        g = Custom(size=100)

        with tempfile.TemporaryDirectory() as tmpdir:
            g.save(os.path.join(tmpdir, "testing-galaxy-custom"))
            g_loaded = sfh.load(os.path.join(tmpdir, "testing-galaxy-custom"))

        self.assertTrue(np.all(g.tau == g_loaded.tau))
        self.assertTrue(np.all(g.rho == g_loaded.rho))
        self.assertTrue(np.all(g.z == g_loaded.z))

    def test_plot(self):
        """Test plotting capabilities"""
        g = sfh.Wagg2022(size=1000)

        it_broke = False
        try:
            g.plot(component="low_alpha_disc", xlim=(-20, 20), ylim=(-20, 20), zlim=(-7, 7), show=False)
            g.plot(coordinates="cylindrical", component="low_alpha_disc", show=False)
        except Exception as e:
            print(e)
            it_broke = True
        self.assertFalse(it_broke)

        it_broke = False
        try:
            g.plot(coordinates="nonsense")
        except ValueError:
            it_broke = True
        self.assertTrue(it_broke)

        plt.close('all')

    def test_indexing(self):
        """Ensure that indexing works correctly (reprs too)"""
        g = sfh.Wagg2022(size=10)
        print(g)

        # make sure it fails for strings
        it_worked = True
        try:
            g["absolute nonsense mate"]
        except ValueError:
            it_worked = False
        self.assertFalse(it_worked)

        inds = [np.random.randint(g.size),
                np.random.randint(g.size, size=4),
                list(np.random.randint(g.size, size=2)),
                slice(0, 7, 3)]

        for ind in inds:
            g_ind = g[ind]
            if isinstance(ind, slice):
                ind = list(range(ind.stop)[ind])
            self.assertTrue(np.all(g.tau[ind] == g_ind.tau))

    def test_spheroidal_dwarf(self):
        """Test the Carina Dwarf SFH class"""
        # first make a bad one
        it_broke = False
        try:
            s = sfh.SpheroidalDwarf(size=1000, J_0_star=1, alpha=1, eta=1, fixed_Z=0.02,
                                    tau_min=5 * u.Gyr, galaxy_age=4 * u.Gyr)
        except ValueError:
            it_broke = True
        self.assertTrue(it_broke)

        # then a carina dwarf
        s = sfh.CarinaDwarf(size=1000, fixed_Z=0.02, tau_min=0 * u.Gyr, galaxy_age=13.5 * u.Gyr)
        self.assertTrue(np.all(s.Z == 0.02))
        self.assertTrue(np.all(s.tau >= 0 * u.Gyr))
        self.assertTrue(np.all(s.tau <= 13.5 * u.Gyr))

    def test_sb15(self):
        """Test the Sanders & Binney (2015) SFH class"""
        s = sfh.SandersBinney2015(size=100, time_bins=1, potential=gp.MilkyWayPotential(version='v2'),
                                  verbose=True)
        # ensure all thick disc stars are older than thin disc stars
        self.assertTrue(np.all(s.tau[s.which_comp == "thick_disc"] >= np.max(s.tau[s.which_comp == "thin_disc"])))

    def test_custom_df(self):
        """Test a custom DF-based SFH class"""
        class SimpleDF(sfh.DistributionFunctionBasedSFH):
            def __init__(self, size, **kwargs):
                super().__init__(size=size, components=["simple"], component_masses=[1], **kwargs)

            def draw_lookback_times(self):
                self._tau = np.full(self.size, 5.0) * u.Gyr
                return self._tau

            def get_metallicity(self):
                self._Z = np.full(self.size, 0.01)
                return self._Z
            
        s = SimpleDF(size=100, potential=gp.MilkyWayPotential(version='v2'), df={
                'type': 'QuasiIsothermal',
                'Rdisk': 3.45,
                'Rsigmar': 7.8,
                'Rsigmaz': 7.8,
                'sigmar0': (48.3*u.km/u.s).decompose(galactic).value,
                'sigmaz0': (30.7*u.km/u.s).decompose(galactic).value,
                'Sigma0': 1.0,
            }, immediately_sample=False)

        s = SimpleDF(size=100, potential=gp.MilkyWayPotential(version='v2'), df=[{
                'type': 'QuasiIsothermal',
                'Rdisk': 3.45,
                'Rsigmar': 7.8,
                'Rsigmaz': 7.8,
                'sigmar0': (48.3*u.km/u.s).decompose(galactic).value,
                'sigmaz0': (30.7*u.km/u.s).decompose(galactic).value,
                'Sigma0': 1.0,
            }], immediately_sample=False)
        s._which_comp = np.array(["simple"] * s.size)
        s.sample()

        self.assertTrue(np.all(s.tau == 5.0 * u.Gyr))



import numpy as np
import unittest
import cogsworth.pop as pop
import cogsworth.sfh as sfh
import cogsworth.obs.observables as obs
import h5py as h5
import os
import pytest
import astropy.units as u
import matplotlib.pyplot as plt
import tempfile


class Test(unittest.TestCase):
    def test_bad_inputs(self):
        """Ensure the class fails with bad input"""
        it_broke = False
        try:
            pop.Population(n_binaries=-100, use_default_BSE_settings=True)
        except ValueError:
            it_broke = True
        self.assertTrue(it_broke)

        it_broke = False
        try:
            pop.Population(n_binaries=0, use_default_BSE_settings=True)
        except ValueError:
            it_broke = True
        self.assertTrue(it_broke)

    def test_io(self):
        """Check that a population can be saved and re-loaded"""

        with tempfile.TemporaryDirectory() as tmpdir:
            # do one with just initial sampling
            p = pop.Population(2, processes=1, sampling_params={"qmin": 0.5}, use_default_BSE_settings=True)
            p.sample_initial_galaxy()
            p.sample_initial_binaries()
            p.save(os.path.join(tmpdir, "testing-pop-io"), overwrite=True)

            p_loaded = pop.load(os.path.join(tmpdir, "testing-pop-io"), 
                                parts=["initial_binaries", "initial_galaxy"])

            self.assertTrue(np.all(p.initial_binaries["mass_1"] == p_loaded.initial_binaries["mass_1"]))

            # again with everything
            p = pop.Population(2, processes=1, bcm_timestep_conditions=[['dtp=100000.0']],
                            sampling_params={"qmin": 0.5}, use_default_BSE_settings=True)
            p.create_population()

            p.save(os.path.join(tmpdir, "testing-pop-io"), overwrite=True)

            p_loaded = pop.load(os.path.join(tmpdir, "testing-pop-io"),
                                parts=["initial_binaries", "initial_galaxy",
                                       "stellar_evolution", "galactic_orbits"])

            self.assertTrue(np.all(p.bpp == p_loaded.bpp))
            self.assertTrue(np.all(p.final_pos == p_loaded.final_pos))
            self.assertTrue(np.all(p.orbits[0].pos == p_loaded.orbits[0].pos))
            self.assertTrue(np.all(p.initial_galaxy.v_R == p_loaded.initial_galaxy.v_R))
            self.assertTrue(p.sampling_params == p_loaded.sampling_params)

            # attempt overwrite without setting flag
            it_broke = False
            try:
                p.save(os.path.join(tmpdir, "testing-pop-io"))
            except FileExistsError:
                it_broke = True
            self.assertTrue(it_broke)

            # attempt overwrite now WITH the flag
            it_broke = False
            try:
                p.save(os.path.join(tmpdir, "testing-pop-io"), overwrite=True)
            except Exception as e:
                print(e)
                it_broke = True
            self.assertFalse(it_broke)

    def test_io_types(self):
        """Ensure that certain variables still have the same type after saving and loading"""
        p = pop.Population(2, processes=1, use_default_BSE_settings=True)
        p.create_population()

        with tempfile.TemporaryDirectory() as tmpdir:
            p.save(os.path.join(tmpdir, "testing-pop-io-types"), overwrite=True)
            p_loaded = pop.load(os.path.join(tmpdir, "testing-pop-io-types"))

        self.assertTrue(type(p.bcm_timestep_conditions) == type(p_loaded.bcm_timestep_conditions))
        self.assertTrue(type(p.bpp_columns) == type(p_loaded.bpp_columns))
        self.assertTrue(type(p.bcm_columns) == type(p_loaded.bcm_columns))

    def test_io_cols(self):
        """Ensure that loading columns is possible"""
        p = pop.Population(2, processes=1, use_default_BSE_settings=True,
                           bpp_columns=["tphys", "mass_1", "mass_2", "sep", "evol_type"],
                           bcm_columns=["tphys", "mass_1", "mass_2", "porb", "ecc", "sep"],
                           bcm_timestep_conditions=[['dtp=0.0']])
        p.create_population()

        with tempfile.TemporaryDirectory() as tmpdir:
            p.save(os.path.join(tmpdir, "testing-pop-io-cols"), overwrite=True)
            p_loaded = pop.load(os.path.join(tmpdir, "testing-pop-io-cols"))

        self.assertTrue(set(p.bpp.columns) == set(p_loaded.bpp.columns))
        self.assertTrue(set(p.bcm.columns) == set(p_loaded.bcm.columns))

    def test_io_versions(self):
        """Check that version mismatches are warned about"""
        p = pop.Population(2, processes=1, use_default_BSE_settings=True)
        p.create_population()

        with tempfile.TemporaryDirectory() as tmpdir:
            p.save(os.path.join(tmpdir, "testing-pop-io-versions"), overwrite=True)

            # load the file and mess with the versions
            with h5.File(os.path.join(tmpdir, "testing-pop-io-versions.h5"), "a") as f:
                f.attrs["cogsworth_version"] = "0.0.0"
                f.attrs["COSMIC_version"] = "0.0.0"
                f.attrs["gala_version"] = "0.0.0"

            with self.assertLogs("cogsworth", level="WARNING") as cm:
                p_loaded = pop.load(os.path.join(tmpdir, "testing-pop-io-versions"))
            self.assertIn("file was saved with", cm.output[0])

    def test_save_complicated_sampling(self):
        """Check that you can save a population with complicated sampling params"""
        p = pop.Population(2, processes=1,
                           sampling_params={
                               "qmin": 0.5,
                               "porb_model": {
                                    "min": 0.15,
                                    "max": 5,
                                    "slope": 0.0
                                }
                           }, use_default_BSE_settings=True)
        p.create_population()

        with tempfile.TemporaryDirectory() as tmpdir:
            p.save(os.path.join(tmpdir, "testing-pop-io-sampling"), overwrite=True)
            p_loaded = pop.load(os.path.join(tmpdir, "testing-pop-io-sampling"), parts=["initial_binaries"])

        # sort columns in both initC to ensure they match
        p._initC = p.initC.reindex(sorted(p.initC.columns), axis=1)
        p_loaded._initC = p_loaded.initC.reindex(sorted(p_loaded.initC.columns), axis=1)

        self.assertTrue(np.all(p.initC == p_loaded.initC))

    def test_lazy_io(self):
        """Check that a population can be saved and re-loaded lazily"""
        p = pop.Population(2, processes=1, bcm_timestep_conditions=[['dtp=100000.0']],
                           sampling_params={"qmin": 0.5}, use_default_BSE_settings=True)
        p.create_population()
        
        with tempfile.TemporaryDirectory() as tmpdir:
            p.save(os.path.join(tmpdir, "testing-lazy-io"), overwrite=True)
            p_loaded = pop.load(os.path.join(tmpdir, "testing-lazy-io"), parts=[])

            # sort columns in both initC to ensure they match
            p._initC = p.initC.reindex(sorted(p.initC.columns), axis=1)
            p_loaded._initC = p_loaded.initC.reindex(sorted(p_loaded.initC.columns), axis=1)

            self.assertTrue(np.all(p.initC == p_loaded.initC))
            self.assertTrue(np.all(p.bpp == p_loaded.bpp))
            self.assertTrue(np.all(p.final_pos == p_loaded.final_pos))
            self.assertTrue(np.all(p.orbits[0].pos == p_loaded.orbits[0].pos))
            self.assertTrue(np.all(p.initial_galaxy.v_R == p_loaded.initial_galaxy.v_R))
            self.assertTrue(p.sampling_params == p_loaded.sampling_params)


    def test_load_no_orbits(self):
        """Check that a population can be saved without orbits, and raises an error if trying to load them"""
        p = pop.Population(2, processes=1, bcm_timestep_conditions=[['dtp=100000.0']],
                           sampling_params={"qmin": 0.5}, use_default_BSE_settings=True)
        p.sample_initial_galaxy()
        p.sample_initial_binaries()
        p.perform_stellar_evolution()
        
        with tempfile.TemporaryDirectory() as tmpdir:
            p.save(os.path.join(tmpdir, "testing-no-orbits"), overwrite=True)
            p_loaded = pop.load(os.path.join(tmpdir, "testing-no-orbits"), parts=[])

            it_broke = False
            try:
                p_loaded.orbits
            except ValueError:
                it_broke = True
            self.assertTrue(it_broke)


    def test_wrong_load_function(self):
        """Check that errors are properly raised when the wrong load function is used"""
        g = sfh.Wagg2022(10000)

        with tempfile.TemporaryDirectory() as tmpdir:
            g.save(os.path.join(tmpdir, "test-sfh-for-load"))

            it_broke = False
            try:
                pop.load(os.path.join(tmpdir, "test-sfh-for-load"))
            except ValueError:
                it_broke = True
            self.assertTrue(it_broke)

            p = pop.Population(2, processes=1, use_default_BSE_settings=True)
            p.create_population()
            p.save(os.path.join(tmpdir, "test-pop-for-load"), overwrite=True)
            it_broke = False
            try:
                sfh.load(os.path.join(tmpdir, "test-pop-for-load"))
            except ValueError:
                it_broke = True
            self.assertTrue(it_broke)

    def test_orbit_storage(self):
        """Test that we can control how orbits are stored"""
        p = pop.Population(20, final_kstar1=[13, 14], processes=1, store_entire_orbits=True,
                           use_default_BSE_settings=True)
        p.create_population()

        first_orbit = p.orbits[0][0] if isinstance(p.orbits[0], list) else p.orbits[0]
        self.assertTrue(first_orbit.shape[0] >= 1)

        p = pop.Population(20, final_kstar1=[13, 14], processes=1, store_entire_orbits=False,
                           use_default_BSE_settings=True)
        p.create_population()

        first_orbit = p.orbits[0][0] if isinstance(p.orbits[0], list) else p.orbits[0]

        self.assertTrue(first_orbit.shape[0] == 1)

    def test_interface(self):
        """Test the interface of this class with the other modules"""
        p = pop.Population(10, processes=2, final_kstar1=[13, 14], store_entire_orbits=False,
                           use_default_BSE_settings=True)
        p.create_population()

        # ensure we get something that disrupts to ensure coverage
        MAX_REPS = 5
        i = 0
        while not p.disrupted.any() and i < MAX_REPS:
            p = pop.Population(10, processes=2, final_kstar1=[13, 14], store_entire_orbits=False,
                               use_default_BSE_settings=True)
            p.create_population()
            i += 1
        if i == MAX_REPS:
            raise ValueError("Couldn't make anything disrupt :/")

        # test that classes can be identified
        self.assertTrue(p.classes.shape[0] == p.n_binaries_match)

        # test that observable table is done right (with or without extinction)
        try:
            p.observables
        except ValueError:
            pass
        p.get_observables(filters=["Gaia_G_EDR3", "Gaia_BP_EDR3", "Gaia_RP_EDR3"],
                          assume_mw_galactocentric=True, ignore_extinction=True)
        obs.get_photometry(filters=["Gaia_G_EDR3", "Gaia_BP_EDR3", "Gaia_RP_EDR3"], final_bpp=p.final_bpp,
                           final_pos=p.final_pos, assume_mw_galactocentric=True, ignore_extinction=True)
        p.observables

        # cheat and make sure at least one binary is bright enough
        p.observables["Gaia_G_EDR3_app_1"].iloc[0] = 18.0
        p.get_gaia_observed_bin_nums(ra="auto", dec="auto")

        it_worked = True
        try:
            p.get_healpix_inds()
        except ValueError:
            it_worked = False
        self.assertFalse(it_worked)

        p.plot_map(ra="auto", dec="auto", coord="C", show=False)
        p.plot_map(ra="auto", dec="auto", coord="G", show=False)
        p.plot_sky_locations(show=False)

        # find a binary that disrupted
        bn = p.bin_nums[p.disrupted][0]
        p.plot_orbit(bn, show=False)
        p.plot_orbit(bn, t_max=0.1 * u.Myr, show=False)
        plt.close("all")

    def test_initial_binaries_replace_initC(self):
        """Test that initial binaries returns initC if present and initial_binaries is not"""

        p = pop.Population(2, processes=1, use_default_BSE_settings=True)
        p.sample_initial_binaries()
        p.sample_initial_galaxy()
        p.perform_stellar_evolution()

        p._initial_binaries = None
        self.assertTrue(p.initial_binaries is p.initC)

    def test_getters(self):
        """Test the property getters"""
        p = pop.Population(2, processes=1, store_entire_orbits=False,
                           bcm_timestep_conditions=[['dtp=1000.0']], use_default_BSE_settings=True)
        p.create_population()

        # test getters from sampling
        p.mass_singles
        p._mass_singles = None
        try:
            p.mass_singles
        except ValueError:
            pass

        p.mass_binaries
        p._mass_binaries = None
        try:
            p.mass_binaries
        except ValueError:
            pass

        p.n_singles_req
        p._n_singles_req = None
        try:
            p.n_singles_req
        except ValueError:
            pass

        p.n_bin_req
        p._n_bin_req = None
        try:
            p.n_bin_req
        except ValueError:
            pass

        p.initial_galaxy
        p._initial_galaxy = None
        try:
            p.initial_galaxy
        except ValueError:
            pass

        # test getters from stellar evolution
        p.bpp
        p._bpp = None
        try:
            p.bpp
        except ValueError:
            pass

        p.bcm
        p._bcm = None
        try:
            p.bcm
        except ValueError:
            pass

        p.initC
        p._initC = None
        try:
            p.initC
        except ValueError:
            pass

        p.kick_info
        p._kick_info = None
        try:
            p.kick_info
        except ValueError:
            pass

        p.create_population()

        # test getters for galactic evolution
        p.orbits
        o = p.orbits.copy()
        p._orbits = None
        try:
            p.orbits
        except ValueError:
            pass
        p._orbits = o

        p.primary_orbits
        p.secondary_orbits

        p._final_vel = None
        p.final_vel

        p.get_final_mw_skycoord()

        p.escaped

    def test_singles_evolution(self):
        """Check everything works well when evolving singles"""
        p = pop.Population(2, processes=1,
                           sampling_params={'keep_singles': True, 'total_mass': 100, 'binfrac_model': 0.0,
                                            'sampling_target': 'total_mass'}, use_default_BSE_settings=True)
        p.create_population(with_timing=False)

        self.assertTrue((p.final_bpp["sep"] == 0.0).all())

    def test_singles_bad_input(self):
        """Test what happens when you mess up single stars"""
        it_failed = True
        p = pop.Population(1, processes=1,
                           sampling_params={'total_mass': 1000, 'sampling_target': 'total_mass',
                                            'binfrac_model': 0.0},
                           use_default_BSE_settings=True)
        try:
            p.sample_initial_binaries()
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

    def test_from_initC(self):
        """Check it can handle only having an initC rather than initial_binaries"""
        p = pop.Population(2, use_default_BSE_settings=True)
        p.sample_initial_binaries()
        p.perform_stellar_evolution()
        p._initial_binaries = None
        p.perform_stellar_evolution()

    def test_none_orbits(self):
        """Ensure final_pos/vel still works when there is an Orbit with None"""
        p = pop.Population(2, use_default_BSE_settings=True)
        p._orbits = [None, None]
        self.assertTrue(p.final_pos[0][0].value == np.inf)
        self.assertTrue(p.final_vel[0][0].value == np.inf)

    @pytest.mark.filterwarnings("ignore:.*duplicate")
    def test_indexing(self):
        """Ensure that indexing works as expected for proper types"""
        p = pop.Population(10, bcm_timestep_conditions=[['dtp=100000.0']], use_default_BSE_settings=True)
        p.create_population()
        inds = [int(np.random.choice(p.bin_nums, replace=False)),
                np.random.choice(p.bin_nums, size=4, replace=False),
                list(np.random.choice(p.bin_nums, size=2, replace=False)),
                slice(0, 7, 3),
                [0, 1, 1, 1, 0]]

        # mock up some data so it tests the indexing
        p._observables = p.final_bpp
        p._classes = p.final_bpp
        p._final_pos = np.zeros(len(p._orbits))
        p._final_vel = np.zeros(len(p._orbits))

        p.copy()

        for ind in inds:
            p_ind = p[ind]
            if isinstance(ind, slice):
                ind = list(range(ind.stop)[ind])
            elif isinstance(ind, int):
                ind = [ind]
            og_m1 = p.final_bpp.loc[ind]["mass_1"].values
            self.assertTrue(np.all(og_m1 == p_ind.final_bpp["mass_1"].values))

    def test_indexing_bad_type(self):
        """Ensure that indexing breaks on bad types (reprs too)"""
        p = pop.Population(10, use_default_BSE_settings=True)
        print(p)
        p.create_population()
        print(p)

        # make sure it fails for strings
        it_worked = True
        try:
            p["absolute nonsense mate"]
        except ValueError:
            it_worked = False
        self.assertFalse(it_worked)

    def test_indexing_bad_bin_num(self):
        """Ensure that indexing breaks on non-existent bin nums"""
        p = pop.Population(10, use_default_BSE_settings=True)
        p.create_population()

        # make sure it fails for bin_nums that don't exist
        it_worked = True
        try:
            p[-42]
        except ValueError:
            it_worked = False
        self.assertFalse(it_worked)

    def test_indexing_booleans(self):
        """Ensure that indexing allows a boolean mask, but breaks on a dodgy version"""
        p = pop.Population(10, use_default_BSE_settings=True)
        p.create_population()

        # make sure it fails for a list of bools of the wrong length
        it_worked = True
        try:
            p[[True, False, True]]
        except AssertionError:
            it_worked = False
        self.assertFalse(it_worked)

        # make sure it works for a proper mask
        it_worked = True
        try:
            p[[True for _ in range(p.n_binaries_match)]]
        except AssertionError:
            it_worked = False
        self.assertTrue(it_worked)

    def test_indexing_mixed_types(self):
        """Don't allow indexing with mixed types"""
        p = pop.Population(10, use_default_BSE_settings=True)
        p.create_population()

        it_worked = True
        try:
            p[[0, "absolute nonsense mate"]]
        except AssertionError:
            it_worked = False
        self.assertFalse(it_worked)

    def test_indexing_loaded_pop(self):
        """Test indexing warns when trying to slice a half-loaded population"""
        p = pop.Population(10, use_default_BSE_settings=True)
        p.perform_stellar_evolution()

        with tempfile.TemporaryDirectory() as tmpdir:
            with h5.File(os.path.join(tmpdir, "DUMMY.h5"), "w") as f:
                f.create_dataset("orbits", data=[])
            p._file = os.path.join(tmpdir, "DUMMY.h5")
            p._orbits = None

            # ensure a warning is raised about missing parts
            with self.assertLogs("cogsworth", level="WARNING") as cm:
                p[:5]
            self.assertIn("You've just masked a population that wasn't fully loaded", cm.output[0])

    def test_evolved_pop(self):
        """Check that the EvolvedPopulation class works as it should"""
        p = pop.Population(10, use_default_BSE_settings=True)
        p.create_population()

        ep = pop.EvolvedPopulation(n_binaries=p.n_binaries_match, mass_singles=p.mass_singles,
                                   mass_binaries=p.mass_binaries, n_singles_req=p.n_singles_req,
                                   n_bin_req=p.n_bin_req, bpp=p.bpp, bcm=p.bcm, initC=p.initC,
                                   kick_info=p.kick_info)

        cant_do_that = False
        try:
            ep.sample_initial_binaries()
        except NotImplementedError:
            cant_do_that = True
        self.assertTrue(cant_do_that)

        cant_do_that = False
        try:
            ep.perform_stellar_evolution()
        except NotImplementedError:
            cant_do_that = True
        self.assertTrue(cant_do_that)

        ep.create_population()

    def test_bin_nums(self):
        """Check that we are creating the correct bin_nums"""
        p = pop.Population(10, use_default_BSE_settings=True)

        # can't get bin_nums before evolution
        it_failed = False
        try:
            p.bin_nums
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

        p.sample_initial_binaries()
        p.bin_nums
        p._bin_nums = None

        p.perform_stellar_evolution()
        p._final_bpp = None
        p.bin_nums
        p._bin_nums = None

        p.final_bpp
        p.bin_nums
        p._bin_nums = None

        initC_bin_nums = p.initC["bin_num"].unique()
        bpp_bin_nums = p.bpp["bin_num"].unique()

        self.assertTrue(np.all(p.bin_nums == initC_bin_nums))
        self.assertTrue(np.all(p.bin_nums == bpp_bin_nums))

        self.assertTrue(len(p) == len(p.bin_nums))

    def test_sampling(self):
        """Ensure that changing sampling parameters actually has an effect"""
        # choose some random q's, ensure samples actually obey changes
        for qmin in np.random.uniform(0, 1, size=10):
            p = pop.Population(1000, sampling_params={"qmin": qmin}, use_default_BSE_settings=True)
            p.sample_initial_binaries()
            q = p._initial_binaries["mass_2"] / p._initial_binaries["mass_1"]
            self.assertTrue(min(q) >= qmin)

    def test_translation(self):
        """Ensure that COSMIC tables are being translated properly"""
        p = pop.Population(10, bcm_timestep_conditions=[['dtp=100000.0']], use_default_BSE_settings=True)
        p.perform_stellar_evolution()
        p.translate_tables(replace_columns=False, label_type="short")

        self.assertTrue(p.bpp["kstar_1"].dtype == np.int64)
        self.assertTrue((p.bpp["kstar_1_str"][p.bpp["kstar_1"] == 1] == "MS").all())

        p.translate_tables(replace_columns=True)
        self.assertFalse(p.bpp["kstar_1"].dtype == np.int64)

    def test_cartoon(self):
        """Ensure that the cartoon plot works"""
        p = pop.Population(10, final_kstar1=[14], use_default_BSE_settings=True)
        p.perform_stellar_evolution()

        for bin_num in p.bin_nums:
            p.plot_cartoon_binary(bin_num, show=False)
        plt.close("all")

    def test_sampling_with_initC(self):
        """Check we can sample from an initC table"""
        p = pop.Population(10, use_default_BSE_settings=True)
        p.perform_stellar_evolution()

        p.sample_initial_binaries(initC=p.initC,
                                  overwrite_initC_settings=True,
                                  reset_sampled_kicks=True)

    def test_legwork_conversion(self):
        """Check construction of LEGWORK sources"""
        p = pop.Population(100, processes=1, use_default_BSE_settings=True)
        p.create_population()

        it_failed = False
        try:
            p.to_legwork_sources(assume_mw_galactocentric=False)
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

        sources = p.to_legwork_sources(assume_mw_galactocentric=True)
        sources.get_merger_time()

    def test_galactic_pool(self):
        """Check that you can create a pool on the fly for galactic evolution"""
        p = pop.Population(10, processes=2, use_default_BSE_settings=True)
        p.sample_initial_binaries()
        p.sample_initial_galaxy()
        p.perform_stellar_evolution()
        p.perform_galactic_evolution()
        self.assertTrue(p.pool is None)

    def test_concat(self):
        """Check that we can concatenate populations"""
        p = pop.Population(10, use_default_BSE_settings=True)
        q = pop.Population(10, use_default_BSE_settings=True)
        p.perform_stellar_evolution()
        q.perform_stellar_evolution()

        r = p + q
        self.assertTrue(len(r) == len(p) + len(q))
        self.assertTrue(len(r.initC["bin_num"].unique()) == len(r))
        self.assertTrue(len(r.initial_galaxy) == len(r))

        self.assertTrue(len(pop.concat(p)) == len(p))
        self.assertTrue(len(sfh.concat(p.initial_galaxy)) == len(p.initial_galaxy))

    def test_concat_wrong_type(self):
        """Check that we can't concatenate with the wrong type"""
        p = pop.Population(10, use_default_BSE_settings=True)
        it_failed = False
        try:
            p + 1
        except AssertionError:
            it_failed = True
        self.assertTrue(it_failed)

    def test_concat_empty(self):
        it_failed = False
        try:
            sfh.concat()
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

        it_failed = False
        try:
            pop.concat()
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

    def test_concat_mismatch(self):
        """Check that we can't concatenate populations with different stuff"""
        p = pop.Population(10, use_default_BSE_settings=True)
        q = pop.Population(10, use_default_BSE_settings=True)
        p.create_population()

        it_failed = False
        try:
            p + q
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

        q.sample_initial_galaxy()
        q._initial_binaries = None
        it_failed = False
        try:
            p + q
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

        q.sample_initial_binaries()
        it_failed = False
        try:
            p + q
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

        q.perform_stellar_evolution()
        it_failed = False
        try:
            p + q
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

    def test_concat_orbits(self):
        """Check that orbits are correctly concatenated"""
        p = pop.Population(25, use_default_BSE_settings=True, final_kstar1=[14], processes=1)
        q = pop.Population(25, use_default_BSE_settings=True, final_kstar1=[14], processes=1)
        p.create_population()
        q.create_population()

        r = p + q

        self.assertTrue(len(r.orbits) == len(p.orbits) + len(q.orbits))
        
        n_p_dis = p.disrupted.sum()
        n_q_dis = q.disrupted.sum()

        # the first len(p) orbits should be from the first len(p) orbits in p
        # the next len(q) orbits should be from the first len(q) orbits in q
        # the next n_p_dis orbits should be from the disrupted orbits in p
        # the next n_q_dis orbits should be from the disrupted orbits in q
        for i in range(len(p)):
            self.assertTrue(np.array_equal(r.orbits[i].pos.xyz.value, p.orbits[i].pos.xyz.value))
        for i in range(len(q)):
            self.assertTrue(np.array_equal(r.orbits[len(p) + i].pos.xyz.value, q.orbits[i].pos.xyz.value))
        for i in range(n_p_dis):
            self.assertTrue(np.array_equal(r.orbits[len(p) + len(q) + i].pos.xyz.value,
                                           p.orbits[len(p) + i].pos.xyz.value))
        for i in range(n_q_dis):
            self.assertTrue(np.array_equal(r.orbits[len(p) + len(q) + n_p_dis + i].pos.xyz.value,
                                           q.orbits[len(q) + i].pos.xyz.value))


    def test_concat_bin_nums_consistent(self):
        """Check that bin_nums are consistent after concatenation"""
        pops = [
            pop.Population(10, use_default_BSE_settings=True, processes=1)
            for _ in range(3)
        ]
        for p in pops:
            p.sample_initial_binaries()
            p.perform_stellar_evolution()

        total = pop.concat(*pops)
        
        # there should be the same number of unique bin_nums in total as the sum of the individuals
        total_unique_bin_nums = total.initC["bin_num"].nunique()
        sum_individual_unique_bin_nums = sum(p.initC["bin_num"].nunique() for p in pops)
        self.assertEqual(total_unique_bin_nums, sum_individual_unique_bin_nums)

    def test_concat_final_pos(self):
        """Check that final_pos is consistent after concatenation"""
        pops = [
            pop.Population(100, final_kstar1=[13, 14], use_default_BSE_settings=True, processes=1,
                           store_entire_orbits=False)
            for _ in range(2)
        ]
        for p in pops:
            p.create_population()
            p.final_pos
            p._orbits = None

        total = pop.concat(*pops)

        # final_pos should have the correct length
        self.assertEqual(len(total.final_pos), sum(len(p.final_pos) for p in pops))

        # final_pos entries should match those from the individual populations, start with bound systems
        # and then the unbound systems like in a normal population
        index = 0
        for p in pops:
            for pos in p.final_pos[:len(p)]:
                self.assertTrue(np.array_equal(total.final_pos[index], pos))
                index += 1
        for p in pops:
            for pos in p.final_pos[len(p):]:
                self.assertTrue(np.array_equal(total.final_pos[index], pos))
                index += 1


    def test_concat_final_pos_bad_input(self):
        """Check that final_pos concatenation raises error when one population lacks final positions"""
        pops = [
            pop.Population(5, use_default_BSE_settings=True, processes=1,
                           store_entire_orbits=False)
            for _ in range(2)
        ]
        for p in pops:
            p.create_population()
            p.final_pos
            p._orbits = None

        pops[-1]._final_pos = None  # simulate not having final positions for one population

        it_failed = False
        try:
            total = pop.concat(*pops)
        except ValueError:
            it_failed = True
        self.assertTrue(it_failed)

    def test_changing_columns(self):
        """Check that a different choice of bpp and bcm columns works"""
        TEST_COLS = ["mass_1", "mass_2", "tphys", "porb", "sep", "ecc", "evol_type"]
        p = pop.Population(10, processes=1, bpp_columns=TEST_COLS, bcm_columns=TEST_COLS,
                           bcm_timestep_conditions=[["mass_1 < 100", 'dtp=100000.0']],
                           use_default_BSE_settings=True)
        p.create_population()

        # bin_num is always added
        TEST_COLS += ["bin_num"]
        self.assertTrue(set(p.bpp.columns) == set(TEST_COLS))
        self.assertTrue(set(p.bcm.columns) == set(TEST_COLS))

    def test_bad_settings(self):
        """Check that passing settings incorrectly raises errors"""
        it_worked = True
        try:
            p = pop.Population(10, use_default_BSE_settings=False, BSE_settings={})
            p.create_population()
        except ValueError:
            it_worked = False
        self.assertFalse(it_worked)

    def test_primary_secondary_pos_vel(self):
        """Check that primary and secondary orbits and can be accessed via properties"""
        p = pop.Population(100, final_kstar1=[13, 14], use_default_BSE_settings=True, processes=1,
                           store_entire_orbits=False)
        p.create_population()

        self.assertTrue(
            np.allclose(p.final_primary_pos, p.final_pos[:len(p)])
        )
        secondary_pos = p.final_pos[:len(p)]
        secondary_pos[p.disrupted] = p.final_pos[len(p):]
        self.assertTrue(
            np.allclose(p.final_secondary_pos, secondary_pos)
        )

        # same for velocities
        self.assertTrue(
            np.allclose(p.final_primary_vel, p.final_vel[:len(p)])
        )
        secondary_vel = p.final_vel[:len(p)]
        secondary_vel[p.disrupted] = p.final_vel[len(p):]
        self.assertTrue(
            np.allclose(p.final_secondary_vel, secondary_vel)
        )
